"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import json
import numpy as np
from numpyencoder import NumpyEncoder

def JSON_from_keras_model(keras_model, output_dir_json):
    model_json = keras_model.to_json()
    model_dict =  json.loads(model_json)

    input_layer_size = 1
    for i in range(1, len(keras_model.input.shape)): #start in idx 1 cause idx 0 represents batch size, so it's None in inference phase
        input_layer_size = input_layer_size * keras_model.input.shape[i]

    model_dict['config']['data_format'] = 'channels_first'
    if (hasattr(layer['config'],'data_format')
        and (layer['config']['data_format'] == 'channels_last') for layer in model_dict['config']['layers']):
        model_dict['config']['data_format'] = 'channels_last'

    if keras_model.layers[0].__class__.__name__ == 'InputLayer':
        start = 0

    else:
        start = 1
        model_dict['config']['layers'][0]['config']['size'] = input_layer_size
        if( model_dict['config']['data_format'] == 'channels_last'):
            if(type(keras_model.input_shape) == list and len(keras_model.input_shape[0]) == 4):
                model_dict['config']['layers'][0]['config']['input_shape'] = [(shape[0], shape[3], shape[1], shape[2]) for shape in keras_model.input_shape]
            elif((type(keras_model.input_shape) is not list) and (len(keras_model.input_shape) == 4)):
                model_dict['config']['layers'][0]['config']['input_shape'] = (keras_model.input_shape[0], keras_model.input_shape[3], keras_model.input_shape[1], keras_model.input_shape[2])
            else:
                model_dict['config']['layers'][0]['config']['input_shape'] = keras_model.input_shape
        else:
            model_dict['config']['layers'][0]['config']['input_shape'] = keras_model.input_shape
        # pretty_json['config']['layers'][0]['config']['dtype'] = "float32"
    i=0
    idx = start
    for layer_json, layer_keras in zip((model_dict['config']['layers'])[start:], keras_model.layers):
        
        layer_size = 1

        if type(layer_keras.output_shape) is list:
            output_shape = layer_keras.output_shape[0]
            input_shape = layer_keras.input_shape[0]
        else:
            output_shape = layer_keras.output_shape
            input_shape = layer_keras.input_shape
        
        if ((model_dict['config']['data_format'] == 'channels_last')):
            if(len(output_shape) == 4):
                output_shape = (output_shape[0], output_shape[3], output_shape[1], output_shape[2])

            if ((type(input_shape) is list) and (len(input_shape[0]) == 4)):
                input_shape = [(shape[0], shape[3], shape[1], shape[2]) for shape in input_shape]
            elif((type(input_shape) is not list) and (len(input_shape) == 4)):
                input_shape = (input_shape[0], input_shape[3], input_shape[1], input_shape[2])

        print(input_shape, output_shape)
        
        for j in range(1, len(output_shape)): #start at idx 1 cause idx 0 represents batch size, so it's None

            layer_size = layer_size * output_shape[j]
        
        if(layer_json['class_name'] == 'UpSampling2D'):
            layer_json['config']['scale'] = layer_json['config']['size']
        
        layer_json['config']['idx'] = idx
        layer_json['config']['size'] = layer_size
        layer_json['config']['input_shape'] = input_shape
        layer_json['config']['output_shape'] = output_shape
        # layer_json['config']['dtype'] = "float32"
        print(layer_json['class_name'], idx)
        layer_json['config']['prev_layer_idx'] = []

        if(('inbound_nodes' in layer_json.keys()) and (layer_json['inbound_nodes'])):
            for prev_layer_json in model_dict['config']['layers'][start:idx]:
                for k in range(len(layer_json['inbound_nodes'][0])):
                    if (prev_layer_json['config']['name'] == layer_json['inbound_nodes'][0][k][0]):
                        layer_json['config']['prev_layer_idx'].append(prev_layer_json['config']['idx']) 
        else:
            layer_json['config']['prev_layer_idx'] = [idx-1]

        if layer_json['class_name'] == 'Dense' or layer_json['class_name'] == 'Conv2D' :
            if layer_json['config']['dtype'] == 'float64':
                weights = np.float64(layer_keras.get_weights()[0])
                if(len(weights.shape) < 3):
                    for i in range(3-len(weights.shape)): 
                        weights = np.expand_dims(weights, axis=-1)
                weights = np.moveaxis(weights, 2, 0)
                if(len(weights.shape) < 4):
                    weights = np.expand_dims(weights, axis=0)
                    
                layer_json['weights'] = weights
                layer_json['biases'] = np.float64(layer_keras.get_weights()[1])

            elif layer_json['config']['dtype'] == 'float32':
                weights = np.float32(layer_keras.get_weights()[0])
                if(len(weights.shape) < 3):
                    for i in range(3-len(weights.shape)): 
                        weights = np.expand_dims(weights, axis=-1)
                weights = np.moveaxis(weights, 2, 0)
                if(len(weights.shape) < 4):
                    weights = np.expand_dims(weights, axis=0)
                
                layer_json['weights'] = weights
                layer_json['biases'] = np.float32(layer_keras.get_weights()[1])

            i = i + 2
        
        if(layer_json['class_name'] == 'BatchNormalization'):
            if layer_json['config']['dtype'] == 'float64':
                layer_json['beta'] = np.float64(layer_keras.get_weights()[1])
                layer_json['gamma'] = np.float64(layer_keras.get_weights()[0])
                layer_json['moving_mean'] = np.float64(layer_keras.get_weights()[2])
                layer_json['moving_var'] = np.float64(layer_keras.get_weights()[3])

            elif layer_json['config']['dtype'] == 'float32':
                layer_json['beta'] = np.float32(layer_keras.get_weights()[1])
                layer_json['gamma'] = np.float32(layer_keras.get_weights()[0])
                layer_json['moving_mean'] = np.float32(layer_keras.get_weights()[2])
                layer_json['moving_var'] = np.float32(layer_keras.get_weights()[3])
        
        idx += 1

    with open(output_dir_json, 'w', encoding='utf-8') as f:
        json.dump(model_dict, f, indent=4, cls = NumpyEncoder)
    f.close()





"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Conv2D_gemm import Conv2D_gemm

import numpy as np
import pystache

class Conv2D_indirect_gemm(Conv2D_gemm):
    """Implements Conv2D using indirect im2col (or im2row) and GeMM"""
   
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def create_ppatches(self):
        if (self.pad_right or self.pad_left or self.pad_bottom or self.pad_top):
            self.input_h_padded = self.input_height + self.pad_top + self.pad_bottom
            self.input_w_padded = self.input_width + self.pad_left + self.pad_right

            start_idx = np.arange(self.kernel_h)[:,None]*self.input_w_padded + np.arange(self.kernel_w)
            c=self.input_h_padded*self.input_w_padded*np.arange(self.input_channels)
            start_idx=(c[:,None]+start_idx.ravel()).reshape((-1,self.kernel_h,self.kernel_w))
            offset_idx = np.arange(self.output_height, step=self.strides)[:,None]*self.input_w_padded + np.arange(self.output_width, step=self.strides)
            idx_padded_input = (start_idx.ravel()[:,None] + offset_idx.ravel()).flatten()

            idx_of_zeros = []
            j_zeros = np.concatenate((np.arange(self.pad_left), np.arange(self.pad_right)+(self.input_w_padded-self.pad_right)))
            i_zeros = np.concatenate((np.arange(self.pad_top), np.arange(self.pad_bottom)+(self.input_h_padded-self.pad_bottom)))
            for c in range(self.input_channels):
                for i in range(self.input_h_padded):
                    for j in range(self.input_w_padded):
                        if (np.isin(i, i_zeros) or np.isin(j, j_zeros)):
                            idx_of_zeros.append(j + self.input_w_padded*(i+self.input_h_padded*c))
            
            idx_padded_input = np.where(np.isin(idx_padded_input, idx_of_zeros), np.nan, idx_padded_input)
            _, idx_padded_input = np.unique(idx_padded_input, return_inverse=True) 
            self.ppatches=np.where(idx_padded_input==self.input_shape, np.nan, idx_padded_input)

        else:
            start_idx = np.arange(self.kernel_h)[:,None]*self.input_width + np.arange(self.kernel_w)
            c=self.input_height*self.input_width*np.arange(self.input_channels)
            start_idx=(c[:,None]+start_idx.ravel()).reshape((-1,self.kernel_h,self.kernel_w))
            offset_idx = np.arange(self.output_height, step=self.strides)[:,None]*self.input_width + np.arange(self.output_width, step=self.strides)
            self.ppatches = (start_idx.ravel()[:,None] + offset_idx.ravel()).flatten()
            
        if ('gemm_nt' or 'gemm_tt') in self.conv_algorithm:
            self.ppatches = self.ppatches.reshape((self.patches_height, self.patches_width)).transpose().flatten()  
  
        
        output_str = self.previous_layer[0].output_str
        
        s = '\n        {'
        for i in range(len(self.ppatches)):
            if np.isnan(self.ppatches[i]):
                s += '&zero, '
            else:
                s += '&'+output_str+'[' + str(int(self.ppatches[i])) + '], '

        s=s[:-2]
        s+='}'

        return s

    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str(self.local_var)

        gemm_code = self.algo_gemm_mapping[self.conv_algorithm](self.nb_filters, self.patches_width, self.patches_height, 'weights_' + self.name + '_' + str("{:02d}".format(self.idx)), 'ppatches_' + self.name + '_' + str("{:02d}".format(self.idx)), "output_"+str(self.path), True)
        mustach_hash['gemm_code'] = gemm_code

        if('cst' not in self.previous_layer[0].output_str):
            mustach_hash['cst'] = True
            mustach_hash['prev_size'] = self.input_channels*self.input_height*self.input_width
            
        with open(self.template_path+'layers/Conv/template_Conv_indirect_gemm.c.tpl', 'r') as template_file:
            template = template_file.read()
        template_file.close()        
        
        return pystache.render(template, mustach_hash)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from .Conv2D import Conv2D

import pystache

class Conv2D_6loops(Conv2D):
    """Implements Conv2D using the six-loops algorithm (direc conv)"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
   
    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['output_str'] = output_str
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str(self.local_var)

        mustach_hash['nb_filters'] = self.nb_filters
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['input_channels'] = self.input_channels
        mustach_hash['kernel_h'] = self.kernel_h
        mustach_hash['kernel_w'] = self.kernel_w
        mustach_hash['strides'] = self.strides
        mustach_hash['dilation_rate'] = self.dilation_rate
        mustach_hash['pad_left'] = self.pad_left
        mustach_hash['pad_top'] = self.pad_top
        mustach_hash['input_height'] = self.input_height
        mustach_hash['input_width'] = self.input_width

        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str(self.local_var,self.idx,'j + '+str(self.output_width)+'*(i + '+str(self.output_height)+'*f)')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True
        
        with open(self.template_path+'layers/Conv/template_Conv_6loops.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""
import onnx
from ..ONNX_importer.create_layer import *
from ...graph.graph_interpretor import tri_topo

def load_onnx(file_to_parse, conv_algorithm):
    #Loading the model and adding value_info if it's not already in it
    if(type(file_to_parse) == str): 
        model = onnx.load(file_to_parse)
    else:
        model = file_to_parse
        
    if (not model.graph.value_info):
        model = onnx.shape_inference.infer_shapes(model)
    
    #The list of layers
    layers = []
    #Dictionnary of the outputs of each layers: {output_name:layer_idx_from_which_the_output_come_from}
    dict_output = {}
    #Dictionnary of the inputs of each layers: {layer_idx_to_which_the_inputs_go:[inputs_name]}
    dict_input = {}
    #Indice of the layer
    idx = 0

    #Creating and adding all the input layers to the list
    layers.append(create_Input_Layer(model.graph.input[0],idx,dict_output))
    idx+=1
    
    #Going through all the nodes to creat the layers and add them to the list
    for node in model.graph.node:
        if(node.op_type == 'BatchNormalization'):
            if(layers[-1].name == 'Conv2D'):
                fuse_BatchNormalization(node, dict_output, model, layers)
            else:
                layers.append(create_BatchNorm(node,idx,dict_input,dict_output,model))
                idx+=1
        elif(node.op_type in layer_type): #If the node is a useful layer, we add it to the list
            if(node.op_type == "Conv"):    
                layers.append(layer_type[node.op_type](node,idx,dict_input,dict_output,model,conv_algorithm))
                idx+=1
            else:
                layers.append(layer_type[node.op_type](node,idx,dict_input,dict_output,model))
                idx+=1
        elif(node.op_type in unused_layers): #If the node is a layer only used in trainning, we do layer_input=layer_output
            unused_layers[node.op_type](node,dict_output,model)
        elif(node.op_type in activation_layers):
            activation_layers[node.op_type](node,dict_output,model,layers)
        else:
            raise TypeError("Error: layer "+node.op_type+" not supported\n")
    
    for layer_idx in dict_input: #Going through all the indices in the list
        layer = layers[layer_idx] #The indice of a layer is the same a it's position in the list
        for input_name in dict_input[layer_idx]: #Going through all the inputs to that layer
            parent = layers[dict_output[input_name]] #Localising that input in the output dictionnary to find the indice of the parent layer
            layer.previous_layer.append(parent)
            parent.next_layer.append(layer)
            
    
    data_type_py = onnx.helper.tensor_dtype_to_np_dtype(model.graph.input[0].type.tensor_type.elem_type)
    
    if data_type_py == 'float64':
        data_type = 'double'

    elif data_type_py == 'float32':
        data_type = 'float'

    elif data_type_py == 'int':
        data_type = 'long int'
    
    #Sorting the graph and adding the road to the layers
    layers, listRoad, maxRoad, dict_cst = tri_topo(layers)
    layers = list(map(lambda x:x.find_output_str(dict_cst), layers))
    data_format = 'channels_first'
    
    print("Finished model initialization.")    
    return layers, data_type, data_type_py, data_format, maxRoad, dict_cst



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
import numpy as np
import pystache

#The layer which compute the general matrix multiplication
#input: weight tesnsor W and bias tensor B, input tensor T. The tensor must be of 2D
#data: alpha and beta constante used in the operation, transpo un tuple saying if the tensor T or W must be transposed before the operation
#output: The result of the operation """alpha*T*W + beta*B"""
class Gemm(Layer):
    def __init__(self, idx, size, alpha, beta, transA, transB, weights, bias, input_shape, output_shape,activation_function):
        super().__init__() 
        self.name = 'Gemm'
        self.idx = idx
        self.size = size
        
        self.alpha = [alpha]
        self.beta =  [beta]
        self.transpo = (transA,transB)
        self.algo_gemm_mapping = {(0,0):self.write_gemm_nn,
                             (0,1):self.write_gemm_nt,
                             (1,1):self.write_gemm_tt,
                             (1,0):self.write_gemm_tn}
        
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]
        if(input_shape):
            print(input_shape)
            if(transA):
                self.input_height = input_shape[3]
                self.input_width = input_shape[2]
            else:
                self.input_height = input_shape[2]
                self.input_width = input_shape[3]
        else:
            self.input_height = 1
            self.input_width = 1
        
        self.weights = np.asarray(weights)
        self.biases = np.asarray(bias)
        self.activation_function = activation_function
        self.nb_weights = self.count_elements_array(self.weights)
        self.nb_biases = self.count_elements_array(self.biases)
        
    #The various ways to compute the operation: 
    
    #None of the tensor ar transposed
    def write_gemm_nn(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_nn.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def write_gemm_nt(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_nt.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def write_gemm_tn(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output',self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_tn.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def write_gemm_tt(self, m, n, k, A, B):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['m'] = m
        mustach_hash['n'] = n
        mustach_hash['k'] = k
        mustach_hash['A'] = A
        mustach_hash['B'] = B
        mustach_hash['activation_function'] = self.activation_function.write_activation_str('sum')
        mustach_hash['alpha'] = self.alpha
        mustach_hash['beta'] = self.beta
        if (self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output_'+str(self.path),self.idx,'i*'+str(self.ldC)+' + j')

            if (self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        with open(self.template_path+'layers/Gemm/template_gemm_tt.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def forward_path_layer(self,input):
        if(self.transpo[0]):
            input = input.reshape(self.input_width,self.input_height).transpose()
        else:
            input = input.reshape(self.input_height,self.input_width)
            
        if(self.transpo[1]):
            self.weights = self.weights.transpose()
        
        return self.activation_function.compute(self.alpha * np.dot(input,self.weights) + self.beta * self.biases)
    
    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['size'] = self.size
        mustach_hash['road'] = self.path

        mustach_hash['patches_size'] = self.output_width*self.output_height
        mustach_hash['gemm_code'] = self.algo_gemm_mapping[self.transpo](self.output_height, self.output_width, self.input_width, 'weights_' + self.name + '_' + str("{:02d}".format(self.idx)), self.previous_layer[0].output_str)

        with open(self.template_path+'layers/Gemm/template_Gemm.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ...Layer import Layer

from abc import abstractmethod
import pystache
import numpy as np

#The class of the Layers which compute operation with broadcast numpy style
#attribut: none
#input: a list of tensor
#output: the resultant tensor
class Broadcast(Layer):
    def __init__(self, idx, size, input_shapes, output_shape,activation_function, constant = None):
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = ''
        self.input_shapes = input_shapes
        
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]
        self.output_channels = output_shape[1]
        self.specific_operator = ''
        self.activation_function = activation_function
        self.constant = constant
        if(constant is not None):
            self.constant_size = self.count_elements_array(self.constant)

    #Go through all the indices and do the operation
    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[k]')

        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width

        start = 0
        if(self.name == 'Maximum'):
            start = 1
            mustach_hash['output_str0'] = self.previous_layer[0].output_str
            mustach_hash['input_width0'] = self.input_shapes[0][3]
            mustach_hash['input_height0'] = self.input_shapes[0][2]
            mustach_hash['input_channels0'] = self.input_shapes[0][1]
            mustach_hash['max'] = True
        elif(self.name == 'Minimum'):
            start = 1
            mustach_hash['output_str0'] = self.previous_layer[0].output_str
            mustach_hash['input_width0'] = self.input_shapes[0][3]
            mustach_hash['input_height0'] = self.input_shapes[0][2]
            mustach_hash['input_channels0'] = self.input_shapes[0][1]
            mustach_hash['min'] = True
        elif(self.name == 'Average'):
            mustach_hash['Average'] = True
            mustach_hash['prev_size'] = len(self.previous_layer)
        else:
            mustach_hash['other'] = True
        
        mustach_hash['broadcast'] = []
        for k in range(start, len(self.previous_layer)):
            previous_dict = {}
            previous_dict['output_str'] = self.previous_layer[k].output_str
            previous_dict['input_width'] = self.input_shapes[k][3]
            previous_dict['input_height'] = self.input_shapes[k][2]
            previous_dict['input_channels'] = self.input_shapes[k][1]
            if(k != len(self.previous_layer) -1):
                previous_dict['operator'] = self.specific_operator
            mustach_hash['broadcast'].append(previous_dict)
        
        if(self.constant is not None):
            constant_dict = {}
            constant_dict['cst_width'] = self.input_shapes[-1][3]
            constant_dict['cst_height'] = self.input_shapes[-1][2]
            constant_dict['cst_channels'] = self.input_shapes[-1][1]
            constant_dict['operator'] = self.specific_operator
            mustach_hash['constant'] = [constant_dict]

        
        with open(self.template_path+'layers/template_Broadcast.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    @abstractmethod
    def forward_path_layer(self, inputs):
        pass


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

#Sort the graph (list of nodes) based on the topological sort
def tri_topo(dnn):
    list_path = []
    #The sorted list to be returned
    list_layers = []
    #the dict stating which layers need to go in a constant
    dict_cst = {}
    for layer in dnn:
        #If the node isn't sorted, we sort it
        if layer.sorted == None:
            parcours_prof_topo(list_layers, layer)
    for layer in list_layers:
        updatepath(layer,list_path)
        to_save(layer,dict_cst)
    max_path = list_path[-2] + 1
    return list_layers, list_path, max_path, dict_cst

#Compute the sorting of a node
def parcours_prof_topo(list_layers,layer):
    #The node is currently being sorted
    layer.sorted = 1
    for nxt_layer in layer.next_layer:
        #The next_layer need to be found and sorted before the current node
        if nxt_layer.sorted == None:
            parcours_prof_topo(list_layers, nxt_layer)
    #The node is sorted
    layer.sorted = 0
    #And is added to the list of sorted nodes
    list_layers.insert(0,layer)


#The function to open a new path
def setNewpath(layer,listCurrentpaths):
    #if the list isn't in the right format, it return -1
    if(len(listCurrentpaths)%2!=0):
        print("Error: listCurrentpaths must be in format:[name_path, is_path_closed, ...]")
        return -1
    else:
        N = len(listCurrentpaths)//2
        #check if there is a previously closed path avialable 
        for i in range(N):
            #is the path is closed, we open it 
            if (listCurrentpaths[2*i+1]==1):
                layer.path = listCurrentpaths[2*i]
                listCurrentpaths[2*i+1] = 0
                break
        #if there is no path at all, we create the first one
        if(N == 0):
            listCurrentpaths.append(0)
            listCurrentpaths.append(0)
            layer.path = 0
        #if no previously closed path can fit are avialable, we create a new one
        elif(layer.path == None):
            listCurrentpaths.append(listCurrentpaths[-2]+1)
            listCurrentpaths.append(0)
            layer.path = listCurrentpaths[-2]
        
#give a layer the right path    
def updatepath(layer,listCurrentpaths):
    if(layer.previous_layer == []):
        #if the layer has no previous one, we creat a new path
        setNewpath(layer,listCurrentpaths)

    given = False
    #every next_layer need to have a path 
    for nxt_layer in layer.next_layer:
        if ((len(nxt_layer.previous_layer)==1) and (not given)):
            #if the next layer only have one prev_layer, and the path hadn't already be given, it receive the same path
            nxt_layer.path = layer.path
            given = True
        elif(nxt_layer.path != None):
            #if the layer already has a path, we do nothing
            pass
        elif(nxt_layer == layer.next_layer[0]):
            #by convention, if it's the first child amongst other, it receive the same path as the father
            nxt_layer.path = layer.path
            given = True
        else:
            #in any other case, the next layer receive a new path
            setNewpath(nxt_layer,listCurrentpaths)
    
    #if the path isn't given, it is closed
    if (given == False):
        listCurrentpaths[layer.path*2 + 1] = 1

#Function creating the dict {idx_layer:idx_cst} saying if a layer must be stored
def to_save(layer,dict_cst):
    for parent in layer.previous_layer:
        if(parent in dict_cst):
            #if the previous_layer are in the dict, we add one to the number of next_layer already "taken care of"
            parent.sorted+=1

    if((len(layer.next_layer)>1)):
        #if the layer has more than one child, it must be stored.
        if(len(dict_cst) == 0):
            dict_cst[layer] = 1 #if the dict is empty, we create the first cst

        else:
            given = False
            #Going through the dict, starting from the end (the opened cst are at the end of the dict)
            for i in range(len(dict_cst)-1,-1,-1): 
                #extracting the layer at the i-th position 
                past_layer = list(dict_cst.keys())[i] 
                #if the layer is complete, we can re-use the same constant
                if (past_layer.sorted == len(past_layer.next_layer)):  
                    past_layer.sorted = 0
                    dict_cst[layer] = dict_cst[past_layer]
                    given = True
                    break
            if not given:# if no constant have been attributed, we create a new one
                dict_cst[layer] = list(dict_cst.values())[-1] + 1


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import json
from itertools import islice
import numpy as np

from ...graph import graph_interpretor

from ...code_generator import (
    AveragePooling2D, MaxPooling2D,
    Conv2D_6loops, Conv2D_std_gemm, Conv2D_indirect_gemm,
    Constant_Pad,
    Add, Multiply, Subtract, Maximum, Minimum, Average,
    ResizeCubic, ResizeLinear, ResizeNearest,
    Concatenate, InputLayer, Dense, Softmax,  Dot, Flatten, BatchNormalization,
    Linear, ReLu, Sigmoid, TanH, LeakyReLu
)

def create_actv_function_obj(activation_str):

        if activation_str == 'sigmoid':
            return Sigmoid()
        elif activation_str == 'relu':
            return ReLu()
        elif activation_str == 'tanh':
            return TanH()
        elif activation_str == 'linear':
            return Linear()
        elif activation_str == 'leaky_relu':
            return LeakyReLu(0.2)

def create_conv2d_obj(algorithm, **kwargs):
       
    if '6loops' in algorithm:
        return Conv2D_6loops(**kwargs)

    elif 'std_gemm' in algorithm:
        return Conv2D_std_gemm(**kwargs)   

    elif 'indirect_gemm' in algorithm:
        return Conv2D_indirect_gemm(**kwargs)

def create_resize_obj(mode, **kwargs):

    if mode == 'bicubic':
        return ResizeCubic(**kwargs)
    
    elif mode == 'bilinear':
        return ResizeLinear(**kwargs)
    
    else:
        return ResizeNearest(**kwargs)

def load_json(file_to_parse, conv_algorithm):
        
    file = open(file_to_parse, 'r')
    model = json.load(file)
    data_format = model['config']['data_format'] 

    data_type = model['config']['layers'][0]['config']['dtype']

    if data_type == 'float64':
        data_type = 'double'
        data_type_py = np.float64

    elif data_type == 'float32':
        data_type = 'float'
        data_type_py = np.float32

    elif data_type == 'int':
        data_type = 'long int'
        data_type_py = np.int32

    layers = []

    l_temp = InputLayer(0, model['config']['layers'][0]['config']['size'], model['config']['layers'][0]['config']['input_shape'], data_format)

    layers.append(l_temp)

    nb_softmax_layers = 0

    for idx, layer in list(islice(enumerate(model['config']['layers']), 1, None)):
        add_softmax_layer = False
        
        idx += nb_softmax_layers

        if 'activation' in layer['config']:
            if layer['config']['activation'] == 'softmax':
                layer['config']['activation'] = 'linear'
                add_softmax_layer = True
    
        if layer['class_name'] == 'Dense':
            current_layer = Dense(idx=idx,
                                    size=layer['config']['units'],
                                    weights=np.array(data_type_py(layer['weights'])),
                                    biases=data_type_py(layer['biases']),
                                    activation_function=create_actv_function_obj(layer['config']['activation']))
        
        elif layer['class_name'] == 'Conv2D':
            current_layer = create_conv2d_obj(algorithm = conv_algorithm,
                                                conv_algorithm = conv_algorithm,
                                                idx = idx,
                                                size = layer['config']['size'],
                                                padding = layer['config']['padding'],
                                                strides = layer['config']['strides'][0],
                                                kernel_h = layer['config']['kernel_size'][0],
                                                kernel_w = layer['config']['kernel_size'][1],
                                                dilation_rate = layer['config']['dilation_rate'][0],
                                                nb_filters = layer['config']['filters'],
                                                input_shape = layer['config']['input_shape'],
                                                output_shape = layer['config']['output_shape'],
                                                weights = np.array(data_type_py(layer['weights'])), 
                                                biases = data_type_py(layer['biases']),
                                                activation_function = create_actv_function_obj(layer['config']['activation']))
        
        elif layer['class_name'] == 'AveragePooling2D':
            current_layer = AveragePooling2D(idx = idx,
                                                size = layer['config']['size'],
                                                padding = layer['config']['padding'],
                                                strides = layer['config']['strides'][0],
                                                pool_size = layer['config']['pool_size'][0],
                                                input_shape = layer['config']['input_shape'],
                                                output_shape = layer['config']['output_shape'],
                                                activation_function = Linear())
        
        elif layer['class_name'] == 'MaxPooling2D':
            current_layer = MaxPooling2D(idx = idx,
                                            size = layer['config']['size'],
                                            padding = layer['config']['padding'],
                                            strides = layer['config']['strides'][0],
                                            pool_size = layer['config']['pool_size'][0],
                                            input_shape = layer['config']['input_shape'],
                                            output_shape = layer['config']['output_shape'],
                                            activation_function = Linear())
        
        elif layer['class_name'] == 'Flatten':
            current_layer = Flatten(idx = idx,
                                            size = layer['config']['size'],
                                            input_shape = layer['config']['input_shape'],
                                            data_format = data_format)
        
        elif layer['class_name'] == 'Add':
            current_layer = Add(idx = layer['config']['idx'],
                                size = layer['config']['size'],
                                input_shapes = layer['config']['input_shape'],
                                output_shape = layer['config']['output_shape'],
                                activation_function = Linear())

        elif layer['class_name'] == 'Multiply':
            current_layer = Multiply(idx = layer['config']['idx'],
                                        size = layer['config']['size'], 
                                        input_shapes = layer['config']['input_shape'], 
                                        output_shape = layer['config']['output_shape'],
                                        activation_function= Linear())
        
        elif layer['class_name'] == 'Subtract':
            current_layer = Subtract(idx = layer['config']['idx'], 
                                        size = layer['config']['size'], 
                                        input_shapes = layer['config']['input_shape'], 
                                        output_shape = layer['config']['output_shape'],
                                        activation_function= Linear())
        
        elif layer['class_name'] == 'Concatenate':
            axis = layer['config']['axis']
            if(data_format == 'channels_last'):
                if(axis == 3):
                    axis = 1
                else:
                    axis = axis + 1
            current_layer = Concatenate(idx = layer['config']['idx'], 
                                        size = layer['config']['size'],
                                        axis = axis, 
                                        input_shapes = layer['config']['input_shape'], 
                                        output_shape = layer['config']['output_shape'],
                                        activation_function=Linear())
        
        elif layer['class_name'] == 'Maximum':
            current_layer = Maximum(idx = layer['config']['idx'], 
                                        size = layer['config']['size'], 
                                        input_shapes = layer['config']['input_shape'], 
                                        output_shape = layer['config']['output_shape'],
                                        activation_function= Linear())
            
        elif layer['class_name'] == 'Minimum':
            current_layer = Minimum(idx = layer['config']['idx'], 
                                        size = layer['config']['size'], 
                                        input_shapes = layer['config']['input_shape'], 
                                        output_shape = layer['config']['output_shape'],
                                        activation_function= Linear())
        
        elif layer['class_name'] == 'Average':
            current_layer = Average(idx = layer['config']['idx'], 
                                        size = layer['config']['size'], 
                                        input_shapes = layer['config']['input_shape'], 
                                        output_shape = layer['config']['output_shape'],
                                        activation_function= Linear())
        
        elif layer['class_name'] == 'Dot':
            current_layer = Dot(idx = layer['config']['idx'],
                                size = layer['config']['size'],
                                axis= layer['config']['axes'],
                                input_shapes = layer['config']['input_shape'], 
                                output_shape = layer['config']['output_shape'],
                                activation_function = Linear())

        elif layer['class_name'] == 'ZeroPadding2D':
            pads = layer['config']['padding']
            if(type(pads) == int):
                pads = [pads for i in range(8)]
            else:
                if(type(pads[0]) == int):
                    pad_top, pad_bottom = pads[0], pads[0]
                    pad_left, pad_right = pads[1], pads[1]
                else:
                    pad_top, pad_bottom = pads[0][0], pads[0][1]
                    pad_left, pad_right = pads[1][0], pads[1][1]
                pads = [0,0,pad_top,pad_left,0,0,pad_bottom,pad_right]

            current_layer = Constant_Pad(idx = layer['config']['idx'],
                                            size = layer['config']['size'],
                                            pads = pads,
                                            constant_value = 0,
                                            axes = [],
                                            input_shape = layer['config']['input_shape'],
                                            activation_function = Linear())
        
        elif layer['class_name'] == 'BatchNormalization':
            if(layers[-1].name == 'Conv2D'):
                scale = np.array(data_type_py(layer['gamma']))
                bias = np.array(data_type_py(layer['beta']))
                mean = np.array(data_type_py(layer['moving_mean']))
                var = np.array(data_type_py(layer['moving_var']))

                weights = layers[-1].weights
                biases = layers[-1].biases

                for z in range(len(weights[0,0,0,:])):
                    alpha = scale[z]/np.sqrt(var[z] + layer['config']['epsilon'])
                    print(alpha)
                    B = bias[z] - (mean[z]*alpha)
                    weights[:,:,:,z] = alpha*weights[:,:,:,z]
                    biases[z] = alpha*biases[z] + B
    
                layers[-1].weights = weights
                layers[-1].biases = biases

                continue
            else:
                current_layer = BatchNormalization(idx = layer['config']['idx'],
                                                    size = layer['config']['size'],
                                                    input_shape = layer['config']['input_shape'],
                                                    epsilon = layer['config']['epsilon'],
                                                    scale = np.array(data_type_py(layer['gamma'])),
                                                    biases = np.array(data_type_py(layer['beta'])),
                                                    mean = np.array(data_type_py(layer['moving_mean'])),
                                                    var = np.array(data_type_py(layer['moving_var'])),
                                                    activation_function = Linear())
        
        
        elif  'Normalization' in layer['class_name']:
            continue

        elif  'Dropout' in layer['class_name']:
            continue

        elif  layer['class_name'] == 'Reshape':
            continue

        elif  layer['class_name'] == 'Permute':
            continue

        else:
            raise TypeError("Error: layer"+layer['class_name']+" not supported\n")
        
        for i in layer['config']['prev_layer_idx']:
            current_layer.previous_layer.append(layers[i])
            layers[i].next_layer.append(current_layer)

        l_temp = current_layer
        layers.append(current_layer)

        # Separeted method to generate softmax
        if add_softmax_layer:
            nb_softmax_layers += 1
            current_layer = Softmax(idx+1, l_temp.size)
            l_temp.next_layer.append(current_layer)
            current_layer.previous_layer.append(l_temp)
            l_temp = current_layer
            layers.append(current_layer)

    layers, listRoad, maxRoad, dict_cst = graph_interpretor.tri_topo(layers)
    layers = list(map(lambda x:x.find_output_str(dict_cst), layers))
    print("Finished model initialization.")
    file.close()
    return layers, data_type, data_type_py, data_format, maxRoad, dict_cst




"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ...Layer import Layer
from abc import abstractmethod

#The resize Layers
#Only Height and Width are resized
#attribut: a lot of stuff
#input: a tensor to be resized, the desired size or a scale to multiply the size by, the region of interest
#output:the resized tensor
##############  https://onnx.ai/onnx/operators/onnx__Resize.html for more informations
#The strategie is always to go throught the elements, find the coordinate in the original tensor 
# and apply a transformation to the value in the original tensor to find the value to enter in the new tensor
class Resize(Layer):
    
    def __init__(self,idx,size,input_shape,activation_function,axes=[],coordinate_transformation_mode='half_pixel',exclude_outside=0,
                 keep_aspect_ratio_policy='stretch',boolean_resize = None,target_size=[],roi=[],extrapolation_value=0, 
                 nearest_mode = 'round_prefer_floor',cubic_coeff_a = -0.75):
        super().__init__()
        self.idx = idx
        if(type(nearest_mode) == bytes):
            self.nearest_mode = str(nearest_mode)[2:-1]
        else: 
            self.nearest_mode = nearest_mode
        
        self.activation_function = activation_function
        self.cubic_coeff_a = cubic_coeff_a
        self.size = size
        self.axes = axes
        self.exclude_outside = exclude_outside
        self.keep_aspect_ratio_policy = keep_aspect_ratio_policy
        self.roi = roi
        self.name = 'Resize'
        self.extrapolation_value = extrapolation_value
        if(type(coordinate_transformation_mode) == bytes):
            self.coordinate_transformation_mode = str(coordinate_transformation_mode)[2:-1]
        else: 
            self.coordinate_transformation_mode = coordinate_transformation_mode
        self.coordinate_transformation_mode_mapping = {"half_pixel":self.half_pixel, 
                                                       "half_pixel_symmetric":self.half_pixel_symmetric,
                                                       "pytorch_half_pixel":self.pytorch_half_pixel,
                                                       "align_corners":self.align_corners,
                                                       "asymmetric":self.asymmetric,
                                                       "tf_crop_and_resize":self.tf_crop_and_resize}
        
        self.coordinate_transformation_mode_implem_mapping =  {"half_pixel":self.half_pixel_implem, 
                                                                "half_pixel_symmetric":self.half_pixel_symmetric_implem,
                                                                "pytorch_half_pixel":self.pytorch_half_pixel_implem,
                                                                "align_corners":self.align_corners_implem,
                                                                "asymmetric":self.asymmetric_implem,
                                                                "tf_crop_and_resize":self.tf_crop_and_resize_implem}
        
        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]
        self.scale = [1,0,0,0]
        if (boolean_resize):
            self.scale = target_size
            self.output_channels = int(self.input_channels*target_size[1]) #should be equal to input_channels
            self.output_height = int(self.input_height*target_size[2])
            self.output_width = int(self.input_width*target_size[3])
        else:
            self.output_channels = target_size[1] #should be equal to input_channels
            self.output_height = target_size[2]
            self.output_width = target_size[3]
            self.scale[1] = self.output_channels / self.input_channels #should be 1
            self.scale[2] = self.output_height / self.input_height
            self.scale[3] = self.output_width / self.input_width
            
    @abstractmethod
    def forward_path_layer(self,input):
        pass
    
    @abstractmethod
    def generate_inference_code_layer(self):
        pass

    #Defining the several coordinate transformations. cf documentation 
    def half_pixel(self,coord_resized,coord_dim,coord_original):
        s = coord_original + ' = ('+ coord_resized+' + 0.5)/'+ str(self.scale[coord_dim])+' - 0.5;'
        return s

    def half_pixel_implem(self, coordinate, coordinate_dimension):
        return (coordinate + 0.5)/self.scale[coordinate_dimension] - 0.5 
    
    def half_pixel_symmetric(self,coord_resized,coord_dim,coord_original):
        if(coord_dim==2):
            target_length = self.output_height*self.scale[2]
            input_length = self.input_height
        else: 
            target_length = self.output_width*self.scale[3]
            input_length = self.input_width

        s = 'float adjustment = ' + str(int(target_length)) + '/' + str(target_length)  +';\n'
        s += '                float center = ' + str(input_length) + '/2;\n'
        s += '                float offset = center*(1 - adjustment);\n'
        s +='                '+ coord_original + ' = offset + ('+ coord_resized+' + 0.5)/'+ str(self.scale[coord_dim])+' - 0.5;'
        return s
    
    def half_pixel_symmetric_implem(self, coordinate, coordinate_dimension):
        if(coordinate_dimension==2):
            target_length = self.output_height*self.scale[2]
            input_length = self.input_height
        else: 
            target_length = self.output_width*self.scale[3]
            input_length = self.input_width

        adjustment = int(target_length)/target_length
        center = input_length/2
        offset = center*(1 - adjustment)
        return offset + (coordinate + 0.5)/self.scale[coordinate_dimension] - 0.5 

    def pytorch_half_pixel(self,coord_resized,coord_dim,coord_original):
        if(coord_dim==2):
            target_length = self.output_height
        else: 
            target_length = self.output_width
        
        s = coord_original + ' = '
        if (target_length > 1):
            s += '('+ coord_resized+' + 0.5)/'+ str(self.scale[coord_dim])+' - 0.5;'
        else:
            s += '0;' 
        return s
    
    def pytorch_half_pixel_implem(self, coordinate, coordinate_dimension):
        if(coordinate_dimension==2):
            target_length = self.output_height
        else: 
            target_length = self.output_width
        
        if(target_length > 1):
            return (coordinate + 0.5)/self.scale[coordinate_dimension] - 0.5 
        else:
            return 0

    def align_corners(self,coord_resized,coord_dim,coord_original):
        if(coord_dim==2):
            length_original = self.input_height
            length_resized = self.output_height
        else: 
            length_original = self.input_width
            length_resized = self.output_width
            
        s = coord_original + ' = ' +coord_resized+'*(' + str(length_original)+' - 1)/(' + str(length_resized)+' - 1);'
        return s
    
    def align_corners_implem(self, coordinate, coordinate_dimension):
        if(coordinate_dimension==2):
            length_original = self.input_height
            length_resized = self.output_height
        else: 
            length_original = self.input_width
            length_resized = self.output_width
        
        return coordinate * (length_original - 1)/(length_resized - 1)
    
    def asymmetric(self,coord_resized,coord_dim,coord_original):
        s = coord_original + ' = '+ coord_resized+'/'+ str(self.scale[coord_dim]) +';'
        return s
    
    def asymmetric_implem(self, coordinate, coordinate_dimension):
        return coordinate/self.scale[coordinate_dimension]
    
    def tf_crop_and_resize(self,coord_resized,coord_dim,coord_original):
        if(coord_dim==2):
            length_original = self.input_height
            length_resized = self.output_height
            start = self.roi[2]
            end = self.roi[6]
        else: 
            length_original = self.input_width
            length_resized = self.output_width
            start = self.roi[3]
            end = self.roi[7]
        
        s = coord_original + ' = ' 
        if(length_resized > 1):
            s+= str(start) + '*('+ str(length_original)+' - 1) + '+ coord_resized+'*(' +str(end)+' - '+str(start)+')*('+str(length_original)+' - 1)/(' + str(length_resized)+' - 1);'
        else:
            s+= '0.5*(' +str(end)+' - '+str(start)+')*('+str(length_original)+' - 1);\n'
        s+= '                if(('+coord_original+' < 0) || ('+coord_original+' > '+str(length_original)+'){'+coord_original+' = '+ str(self.extrapolation_value)+'}'
        return s

    def tf_crop_and_resize_implem(self, coordinate, coordinate_dimension):
        if(coordinate_dimension==2):
            length_original = self.input_height
            length_resized = self.output_height
            start = self.roi[2]
            end = self.roi[6]
        else: 
            length_original = self.input_width
            length_resized = self.output_width
            start = self.roi[3]
            end = self.roi[7]

        if(length_resized > 1):
            coordinate_original =  start*(length_original - 1) + coordinate*(end - start)*(length_original -1)/(length_resized - 1)
        else:
            coordinate_original = 0.5*(start + end)*(length_original - 1)

        if(coordinate_original >= length_original or coordinate_original < 0):
            return self.extrapolation_value
        else:
            return coordinate_original



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""


import numpy as np
import onnx

from ...code_generator.layers import (
    AveragePooling2D, MaxPooling2D,
    Conv2D_6loops, Conv2D_std_gemm, Conv2D_indirect_gemm,
    Edge_pad, Wrap_pad, Reflect_pad, Constant_Pad,
    Add, Multiply, Subtract, Divide, Maximum, Minimum, Average,
    ResizeCubic, ResizeLinear, ResizeNearest,
    Concatenate, InputLayer, Softmax,  Dot, Gather, Gemm, MatMul, Add_Bias, BatchNormalization
)

from ...code_generator.activation_functions import Linear, ReLu, Sigmoid, TanH, Clip, Exponential, Logarithm, LeakyReLu

###### Utility functions ######

def create_conv2d_obj(algorithm, **kwargs):
       
    if '6loops' in algorithm:
        return Conv2D_6loops(**kwargs)

    elif 'std_gemm' in algorithm:
        return Conv2D_std_gemm(**kwargs)   

    elif 'indirect_gemm' in algorithm:
        return Conv2D_indirect_gemm(**kwargs)
        

def create_resize_obj(mode, **kwargs):
    if mode == b'nearest':
        return ResizeNearest(**kwargs)
    
    elif mode == b'linear':
        return ResizeLinear(**kwargs)
    
    elif mode == b'cubic':
        return ResizeCubic(**kwargs)

def create_pad_obj(mode, **kwargs):
    if mode == b'constant':
        return Constant_Pad(**kwargs)
    elif mode == b'edge':
        return Edge_pad(**kwargs)
    elif mode == b'wrap':
        return Wrap_pad(**kwargs)
    elif mode == b'reflect':
        return Reflect_pad(**kwargs)

#Go find the constant named initialzer_name in model(an onnx model)
def look_for_initializer(initializer_name,model):
    if (initializer_name == ''):
        return []
    for initializer in model.graph.initializer:
        if (initializer.name == initializer_name):
            return initializer
    return []


#Return a dictonnary with {attibut: attribut_values}
def extract_attribut(node):
    attributes = {}
    for attribute in node.attribute:
        attributes[attribute.name] = onnx.helper.get_attribute_value(attribute)
    return attributes

#Find the shape of shape_name in model (an onnx model)
#Depend of the carac of the model. Need value_info in the model
def get_shape(shape_name,model):
    shape = []
    for info in model.graph.value_info:
        if(shape_name == info.name):
            shape = [info.type.tensor_type.shape.dim[i].dim_value for i in range(len(info.type.tensor_type.shape.dim))]
    for input in model.graph.input:
        if(shape_name == input.name):
            shape = [input.type.tensor_type.shape.dim[i].dim_value for i in range(len(input.type.tensor_type.shape.dim))]
    for output in model.graph.output:
        if(shape_name == output.name):
            shape = [output.type.tensor_type.shape.dim[i].dim_value for i in range(len(output.type.tensor_type.shape.dim))]
    if (shape and len(shape)<=4):
        shape = [1 for i in range(4-len(shape))] + shape
    for i in range(len(shape)):
        if shape[i] == 0:
            shape[i] = 1
    return shape

#Return the size of the layer when given the list output_shape
def find_size(output_shape):
    size = 1
    for i in output_shape:
        if i != 0:
            size*=i
    return size


###### Functions to create a Layer ######

#Create an input layers 
def create_Input_Layer(input_layer,idx,dict_output):
        dict_output[input_layer.name] = idx
        output_shape = [input_layer.type.tensor_type.shape.dim[i].dim_value for i in range(len(input_layer.type.tensor_type.shape.dim))]
        size = find_size(output_shape)
        
        return InputLayer(idx,size,output_shape,'channels_first')

#Create a layer Softmax
def create_Softmax(node,idx,dict_input,dict_output,model):
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = node.input
    dict_output[node.output[0]] = idx
    return Softmax(idx = idx,
                            size = size)


#Create a layer Conv
def create_Conv(node,idx,dict_input,dict_output,model,conv_algorithm):
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = [node.input[0]]
    dict_output[node.output[0]] = idx
    initializers = [look_for_initializer(initializer_name,model) for initializer_name in node.input[1:]]
    attributs = extract_attribut(node)
    if ('dilations' not in attributs):
        attributs['dilations'] = [1]
    if ('group' not in attributs):
        attributs['group'] = 1
    if (('auto_pad' not in attributs) or ( attributs['auto_pad'].decode() == 'NOTSET')):
        attributs['auto_pad'] = attributs['pads']
    else:
        attributs['auto_pad'] = attributs['auto_pad'].decode()
    if ('strides' not in attributs):
        attributs['strides'] = [1,1]
        
    if len(initializers)==2:
        biases = onnx.numpy_helper.to_array(initializers[1])
    else:
        biases = np.zeros(output_shape[1])
        
    return create_conv2d_obj(algorithm=conv_algorithm,
                                conv_algorithm=conv_algorithm,
                                idx= idx,
                                size= size,
                                padding= attributs['auto_pad'],
                                strides= attributs['strides'][0],
                                kernel_h= attributs['kernel_shape'][0], 
                                kernel_w= attributs['kernel_shape'][1], 
                                dilation_rate= attributs['dilations'][0], 
                                nb_filters= initializers[0].dims[0],
                                input_shape= input_shape, 
                                output_shape= output_shape,
                                weights= np.moveaxis(onnx.numpy_helper.to_array(initializers[0]), 0,3),
                                biases= biases,
                                activation_function= Linear())
    

#Create a layer Concat
def create_Concat(node,idx,dict_input,dict_output,model):
    input_shapes =[]
    for input in node.input:
        input_shapes.append(get_shape(input,model))
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = node.input
    dict_output[node.output[0]] = idx
    attributs = extract_attribut(node)
    return Concatenate(idx,
                        size, 
                        attributs['axis'],
                        input_shapes,
                        output_shape,
                        activation_function= Linear())
    
#Create a layer Resize
def create_Resize(node,idx,dict_input,dict_output,model):
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = [node.input[0]]
    dict_output[node.output[0]] = idx
    initializers = [look_for_initializer(initializer_name,model) for initializer_name in node.input[1:]]
    attributs = extract_attribut(node)
    if ('axes' not in attributs):
        attributs['axes'] = []
    if('coordinate_transformation_mode' not in attributs):
        attributs['coordinate_transformation_mode'] = 'half_pixel'
    if('exclude_outside' not in attributs):
        attributs['exclude_outside'] = 0
    if('keep_aspect_ratio_policy' not in attributs):
        attributs['keep_aspect_ratio_policy'] = 'stretch'
    if('extrapolation_value' not in attributs):
        attributs['extrapolation_value'] = 0.0
    if('nearest_mode' not in attributs):
        attributs['nearest_mode'] = 'round_prefer_floor'
    if('cubic_coeff_a' not in attributs):
        attributs['cubic_coeff_a'] = -0.75
        
    if(initializers[2]):
        #target size
        targ_size = onnx.numpy_helper.to_array(initializers[2])
        bool_resize = 0
    else:
        #scale
        targ_size = onnx.numpy_helper.to_array(initializers[1])
        bool_resize = 1
        
    if(initializers[0]):
        roi = onnx.numpy_helper.to_array(initializers[0])
    else:
        roi = initializers[0]
    
    return create_resize_obj(mode = attributs['mode'],
                             idx = idx,
                             size = size,
                             input_shape = input_shape,
                             axes = attributs['axes'],
                             coordinate_transformation_mode = attributs['coordinate_transformation_mode'],
                             exclude_outside = attributs['exclude_outside'],
                             keep_aspect_ratio_policy = attributs['keep_aspect_ratio_policy'],
                             target_size = targ_size,
                             boolean_resize = bool_resize,
                             roi = roi,
                             extrapolation_value = attributs['extrapolation_value'],
                             nearest_mode = attributs['nearest_mode'],
                             cubic_coeff_a=attributs['cubic_coeff_a'],
                             activation_function = Linear())
    
#create a layer Pad
def create_Pad(node,idx,dict_input,dict_output,model):
    axes = []
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = [node.input[0]]
    dict_output[node.output[0]] = idx
    initializers = [look_for_initializer(initializer_name,model) for initializer_name in node.input[1:]]
    attributs = extract_attribut(node)
    if(len(initializers)==2):
        initializers.append([])
    if(initializers[2]):
        axes = onnx.numpy_helper.to_array(initializers[2])
    return create_pad_obj(mode = attributs['mode'],
                          idx = idx,
                          size = size,
                          pads = onnx.numpy_helper.to_array(initializers[0]),
                          constant_value = onnx.numpy_helper.to_array(initializers[1]),
                          axes = axes,
                          input_shape = input_shape,
                          activation_function = Linear())


#create a layer Gather
def create_Gather(node,idx,dict_input,dict_output,model):
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = [node.input[0]]
    dict_output[node.output[0]] = idx
    attributs = extract_attribut(node)
    initializer = look_for_initializer(node.input[1],model)
    return Gather(idx = idx,
                    size = size,
                    axis = attributs['axis'],
                    indices = onnx.numpy_helper.to_array(initializer),
                    input_shape = input_shape,
                    output_shape = output_shape,
                    activation_function = Linear())

#create a layer Gemm
def create_Gemm(node,idx,dict_input,dict_output,model):
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = [node.input[0]]
    dict_output[node.output[0]] = idx
    attributs = extract_attribut(node)
    B_tensor = look_for_initializer(node.input[1],model)
    if(len(node.input) == 3):
        C_tensor = look_for_initializer(node.input[2],model)
    else:
        C_tensor = 0
    if('transA' not in attributs):
        attributs['transA'] = 0
    if('transB' not in attributs):
        attributs['transB'] = 0
    if('alpha' not in attributs):
        attributs['alpha'] = 1.0
    if('beta' not in attributs):
        attributs['beta'] = 1.0
    return Gemm(idx = idx,
                size = size,
                alpha = attributs['alpha'],
                beta = attributs['beta'],
                transA = attributs['transA'],
                transB = attributs['transB'],
                weights = onnx.numpy_helper.to_array(B_tensor),
                bias = onnx.numpy_helper.to_array(C_tensor),
                input_shape = input_shape,
                output_shape = output_shape,
                activation_function = Linear())

def create_MatMul(node,idx,dict_input,dict_output,model):
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_output[node.output[0]] = idx
    right_tensor = look_for_initializer(node.input[0],model)
    left_tensor = look_for_initializer(node.input[1],model)
    if(left_tensor or right_tensor):
        if(right_tensor and not left_tensor):
            #the weigth is the right tensor:  MatMul(W,T)
            side = True
            weights = onnx.numpy_helper.to_array(right_tensor)
            weights = np.reshape(weights, (get_shape(node.input[1],model)[-1],1,1,output_shape[-1]))
            weights = np.moveaxis(weights, 0,3)
            dict_input[idx] = [node.input[1]]
            input_shape = get_shape(node.input[1],model)
        if(left_tensor and not right_tensor):
            #the weigth is the right tensor:  MatMul(W,T)
            side = False
            weights = onnx.numpy_helper.to_array(left_tensor)
            weights = np.reshape(weights, (1,1,get_shape(node.input[0],model)[-1],output_shape[-1]))
            dict_input[idx] = [node.input[0]]
            input_shape = get_shape(node.input[0],model)
        return MatMul(idx = idx,
                        size = size,
                        input_shape = input_shape,
                        weights = weights,
                        side = side,
                        activation_function = Linear())
    else:
        dict_input[idx] = node.input
        input_shapes =[]
        for input in node.input:
            input_shapes.append(get_shape(input,model))
        # to check
        return Dot(idx = idx,
                    size = size,
                    axis = [-1,-2],
                    input_shapes = input_shapes,
                    output_shape = output_shape,
                    activation_function = Linear())

def create_BatchNorm(node,idx,dict_input,dict_output,model):
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = [node.input[0]]
    dict_output[node.output[0]] = idx
    attributs = extract_attribut(node)
    
    if('epsilon' not in attributs):
        attributs['epsilon'] = 1e-05
    
    scale = look_for_initializer(node.input[1],model)
    biases = look_for_initializer(node.input[2],model)
    mean = look_for_initializer(node.input[3],model)
    var = look_for_initializer(node.input[4],model)

    return BatchNormalization(idx = idx,
                              size = size,
                              input_shape = output_shape,
                              epsilon = attributs['epsilon'],
                              scale = onnx.numpy_helper.to_array(scale),
                              biases = onnx.numpy_helper.to_array(biases),
                              mean = onnx.numpy_helper.to_array(mean),
                              var = onnx.numpy_helper.to_array(var),
                              activation_function = Linear())



### Pooling layers ###

#Create a layer MaxPool
def create_MaxPool(node,idx,dict_input,dict_output,model):
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = node.input
    dict_output[node.output[0]] = idx
    attributs = extract_attribut(node)
    if ('dilations' not in attributs):
        attributs['dilations'] = [1]
    if (('auto_pad' not in attributs) or ( attributs['auto_pad'] == 'NOTSET')):
        attributs['auto_pad'] = attributs['pads']
    if ('strides' not in attributs):
        attributs['strides'] = [1,1]
    return MaxPooling2D(idx = idx,
                        size = size,
                        padding =attributs['auto_pad'],
                        strides = attributs['strides'][0],
                        pool_size = attributs['kernel_shape'][0],
                        input_shape = input_shape,
                        output_shape = output_shape,
                        activation_function = Linear())

#cerate a layer AveragePool
def create_AveragePool(node,idx,dict_input,dict_output,model):
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = [node.input[0]]
    dict_output[node.output[0]] = idx
    attributs = extract_attribut(node)
    if ('dilations' not in attributs):
        attributs['dilations'] = [1]
    if (('auto_pad' not in attributs) or ( attributs['auto_pad'] == 'NOTSET')):
        attributs['auto_pad'] = attributs['pads']
    if ('strides' not in attributs):
        attributs['strides'] = [1,1]
    return AveragePooling2D(idx=idx,
                            size=size, 
                            padding=attributs['auto_pad'],
                            strides=attributs['strides'][0], 
                            pool_size=attributs['kernel_shape'][0], 
                            input_shape=input_shape,
                            output_shape=output_shape,
                            activation_function = Linear())

#Create a layer GlobalAveragePool
def create_GlobalAveragePool(node,idx,dict_input,dict_output,model):
    input_shape = get_shape(node.input[0],model)
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = node.input
    dict_output[node.output[0]] = idx
    return AveragePooling2D(idx = idx,
                            size = size,
                            padding = [0,0,0,0],
                            strides = 0,
                            pool_size = input_shape[2],
                            input_shape = input_shape,
                            output_shape = output_shape,
                            activation_function = Linear())

### Broadcats layers ###

#create a layer Add_Biass 
##### UNUSED #####
def create_Add_Biass(node,idx,dict_input,dict_output,model):
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_output[node.output[0]] = idx
    right_tensor = look_for_initializer(node.input[0],model)
    left_tensor = look_for_initializer(node.input[1],model)
    if(right_tensor):
        biases = onnx.numpy_helper.to_array(right_tensor)
        dict_input[idx] = [node.input[1]]
    elif(left_tensor):
        biases = onnx.numpy_helper.to_array(left_tensor)
        dict_input[idx] = [node.input[0]]
    return Add_Bias(idx = idx,
                        size = size,
                        biases = biases,
                        activation_function = Linear())

#create a layer Add
def create_Add(node,idx,dict_input,dict_output,model):
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_output[node.output[0]] = idx
    dict_input[idx] = []
    constant = np.zeros(get_shape(node.input[0], model))
    input_shapes =[]
    for input in node.input:
        cst = look_for_initializer(input,model)
        if(cst):
            constant = constant + onnx.numpy_helper.to_array(cst)
        else:
            dict_input[idx].append(input)
            input_shapes.append(get_shape(input,model))
    if(constant.any()):
        if (len(constant.shape)<4):
            for i in range(0,4-len(constant.shape)):
                constant = np.expand_dims(constant,axis=0)
        input_shapes.append(constant.shape)
    else:
        constant = None
    return Add(idx = idx,
                size = size,
                input_shapes = input_shapes,
                output_shape = output_shape,
                activation_function = Linear(),
                constant = constant)
    
#create a layer Div
def create_Div(node,idx,dict_input,dict_output,model):
    input_shapes =[]
    constant = 1
    dict_input[idx] = []
    for input in node.input:
        factor = look_for_initializer(input,model)
        if(factor):
            constant = constant / onnx.numpy_helper.to_array(factor)
        else:
            dict_input[idx].append(input)
            input_shapes.append(get_shape(input,model))
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_output[node.output[0]] = idx
    if((constant == np.ones(constant.shape)).all()):
        constant = None
    else:
        if (len(constant.shape)<4):
            for i in range(0,4-len(constant.shape)):
                constant = np.expand_dims(constant,axis=0)
        input_shapes.append(constant.shape)
    return Divide(idx=idx,
                    size=size,
                    input_shapes=input_shapes,
                    output_shape=output_shape,
                    activation_function= Linear(),
                    constant=constant)


#create a layer Mul
def create_Mul(node,idx,dict_input,dict_output,model):
    input_shapes =[]
    constant = 1
    dict_input[idx] = []
    for input in node.input:
        factor = look_for_initializer(input,model)
        if(factor):
            constant = constant * onnx.numpy_helper.to_array(factor)
        else:
            dict_input[idx].append(input)
            input_shapes.append(get_shape(input,model))
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_output[node.output[0]] = idx
    if((constant == np.ones(constant.shape)).all()):
        constant = None
    else:
        if (len(constant.shape)<4):
            for i in range(0,4-len(constant.shape)):
                constant = np.expand_dims(constant,axis=0)
        input_shapes.append(constant.shape)
    return Multiply(idx=idx,
                    size=size,
                    input_shapes=input_shapes,
                    output_shape=output_shape,
                    activation_function= Linear(),
                    constant=constant)

#create a layer Sub
def create_Sub(node,idx,dict_input,dict_output,model):
    input_shapes =[]
    constant = 0
    dict_input[idx] = []
    for input in node.input:
        factor = look_for_initializer(input,model)
        if(factor):
            constant = constant - onnx.numpy_helper.to_array(factor)
        else:
            dict_input[idx].append(input)
            input_shapes.append(get_shape(input,model))
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_output[node.output[0]] = idx
    if((constant == np.ones(constant.shape)).all()):
        constant = None
    else:
        if (len(constant.shape)<4):
            for i in range(0,4-len(constant.shape)):
                constant = np.expand_dims(constant,axis=0)
        input_shapes.append(constant.shape)
    return Subtract(idx=idx,
                    size=size,
                    input_shapes=input_shapes,
                    output_shape=output_shape,
                    activation_function= Linear(),
                    constant=constant)
    
#create a layer Max
def create_Max(node,idx,dict_input,dict_output,model):
    input_shapes =[]
    for input in node.input:
        input_shapes.append(get_shape(input,model))
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = node.input
    dict_output[node.output[0]] = idx
    return Maximum(idx=idx,
                    size=size,
                    input_shapes=input_shapes,
                    output_shape=output_shape,
                    activation_function= Linear())

#create a layer Min
def create_Min(node,idx,dict_input,dict_output,model):
    input_shapes =[]
    for input in node.input:
        input_shapes.append(get_shape(input,model))
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = node.input
    dict_output[node.output[0]] = idx
    return Minimum(idx=idx,
                    size=size,
                    input_shapes=input_shapes,
                    output_shape=output_shape,
                    activation_function= Linear())

#create a layer Average
def create_Avg(node,idx,dict_input,dict_output,model):
    input_shapes =[]
    for input in node.input:
        input_shapes.append(get_shape(input,model))
    output_shape = get_shape(node.output[0],model)
    size = find_size(output_shape)
    dict_input[idx] = node.input
    dict_output[node.output[0]] = idx
    return Average(idx=idx,
                    size=size,
                    input_shapes=input_shapes,
                    output_shape=output_shape,
                    activation_function= Linear())

###### Dict of all the functions ######
layer_type = {"Softmax":create_Softmax,
         "Conv":create_Conv, 
         "Resize":create_Resize,
         "Pad":create_Pad,
         "Concat":create_Concat,
         "Gather":create_Gather,
         "Gemm":create_Gemm,
         "MatMul":create_MatMul,
         "MaxPool":create_MaxPool,
         "AveragePool":create_AveragePool,
         "GlobalAveragePool":create_GlobalAveragePool,
         "Add":create_Add,
         "Mul":create_Mul,
         "Div":create_Div,
         "Sub":create_Sub,
         "Max":create_Max,
         "Min":create_Min,
         "Mean":create_Avg}


###### Function to deal with the 'non important' layers of the graph ######

#Do the opeartion: Dropout.input = Dropout.output
def bypass(node,dict_output,model):
    dict_output[node.output[0]] = dict_output.pop(node.input[0])

def create_initializer(node,dict_output,model):
    const = model.graph.initializer.add()
    const.name = node.output[0]
    const.data_type = node.attribute[0].t.data_type
    const.raw_data = node.attribute[0].t.raw_data

###### Dict of all the functions ######
unused_layers = {"Dropout":bypass,
                  "Constant":create_initializer,
                  "Unsqueeze":bypass,
                  "Reshape":bypass,
                  "LRN":bypass,
                  "Shape":bypass}

###### Function to fuse to ONNX layers ######

#Fuse the activation layer ReLu with the prior layer
def fuse_ReLu(node,dict_output,model,layers):
    layers[dict_output[node.input[0]]].activation_function = ReLu()
    bypass(node,dict_output,model)

#Fuse the activation layer Tanh with the prior layer
def fuse_Tanh(node,dict_output,model,layers):
    layers[dict_output[node.input[0]]].activation_function = TanH()
    bypass(node,dict_output,model)

#Fuse the activation layer Sigmoide with the prior layer
def fuse_Sigmoid(node,dict_output,model,layers):
    layers[dict_output[node.input[0]]].activation_function = Sigmoid()
    bypass(node,dict_output,model)

#Fuse the activation layer Sigmoide with the prior layer
def fuse_Exp(node,dict_output,model,layers):
    layers[dict_output[node.input[0]]].activation_function = Exponential()
    bypass(node,dict_output,model)

#Fuse the activation layer Sigmoid with the prior layer
def fuse_Log(node,dict_output,model,layers):
    layers[dict_output[node.input[0]]].activation_function = Logarithm()
    bypass(node,dict_output,model)

#Fuse a layer Clip with the prior layer
def fuse_Clip(node,dict_output,model,layers):
    min, max = float('-inf'), float('inf')
    if(node.input[1]):
        min = onnx.numpy_helper.to_array(look_for_initializer(node.input[1],model))[0]
    if(node.input[2]):
        max = onnx.numpy_helper.to_array(look_for_initializer(node.input[2],model))[0]
    layers[dict_output[node.input[0]]].activation_function = Clip(max=max,min=min)
    bypass(node,dict_output,model)

#Fuse the activation layer LeakyRelu with the prior layer
def fuse_LeakyReLu(node,dict_output,model,layers):
    attribut = extract_attribut(node)
    if('alpha' not in attribut):
        attribut['alpha'] = 0.01
    layers[dict_output[node.input[0]]].activation_function = LeakyReLu(alpha=attribut['alpha'])
    bypass(node,dict_output,model)

#Fuse a BatchNormalization layer with the previous Conv2D layer
def fuse_BatchNormalization(node,dict_output,model,layers):
    attributs = extract_attribut(node)
    if('epsilon' not in attributs):
        attributs['epsilon'] = 1e-05
    
    scale = onnx.numpy_helper.to_array(look_for_initializer(node.input[1],model))
    bias = onnx.numpy_helper.to_array(look_for_initializer(node.input[2],model))
    mean = onnx.numpy_helper.to_array(look_for_initializer(node.input[3],model))
    var = onnx.numpy_helper.to_array(look_for_initializer(node.input[4],model))

    weights = layers[dict_output[node.input[0]]].weights
    biases = layers[dict_output[node.input[0]]].biases

    for z in range(len(weights[0,0,0,:])):
        alpha = scale[z]/np.sqrt(var[z] + attributs['epsilon'])
        B = bias[z] - (mean[z]*alpha)
        weights[:,:,:,z] = alpha*weights[:,:,:,z]
        biases[z] = alpha*biases[z] + B
    
    layers[dict_output[node.input[0]]].weights = weights
    layers[dict_output[node.input[0]]].biases = biases

    bypass(node, dict_output, model)

###### Dict of all the functions ######
activation_layers = {"Relu":fuse_ReLu,
                     "Tanh":fuse_Tanh,
                     "Sigmoid":fuse_Sigmoid,
                     "Clip":fuse_Clip,
                     "Exp":fuse_Exp,
                     "Log":fuse_Log,
                     "LeakyRelu":fuse_LeakyReLu}


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
import numpy as np
import pystache

class Add_Bias(Layer):

    def __init__(self, idx, size, biases,activation_function):
        
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'Add_Biass'
        self.biases = np.asarray(biases)
        self.nb_biases = self.count_elements_array(self.biases)
        self.activation_function = activation_function
    
    #Go through all the indices and do the operation
    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str#if the value is in a road or saved eslewhere

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['output_str'] = output_str
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output_'+str(self.path)+'[i]')

        if(self.activation_function.name == 'linear'):
                mustach_hash['linear'] = True

        if(self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('output_'+str(self.path)+'[i]',self.idx,'i')

        with open(self.template_path+'layers/template_AddBiase.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)

    def forward_path_layer(self, input):
        input = input.reshape(self.previous_layer[0].size)

        return self.activation_function.compute(input + self.biases)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import argparse

from ..cli_acetone import cli_acetone
from ..cli_compare import cli_compare

def acetone_generate():
    parser = argparse.ArgumentParser(description='C code generator for neural networks')

    parser.add_argument("model_file", help="Input file that describes the neural network model")
    parser.add_argument("function_name", help="Name of the generated function")
    parser.add_argument("nb_tests", help="Number of inferences process to run")
    parser.add_argument("conv_algorithm", help="Algorithm to be used in convolutional layer. Default is indirect im2col with GeMM")
    parser.add_argument("output_dir", help="Output directory where generated files will be written")
    parser.add_argument("test_dataset_file", nargs='?', default=None, help="Input file that contains test data")
    parser.add_argument("normalize", nargs='?', default=False, help="Boolean saying if the inputs and outputs needs to be normalized. Only used when the file is in NNET representation")

    args = parser.parse_args()

    cli_acetone(model_file=args.model_file, 
                function_name=args.function_name, 
                nb_tests=args.nb_tests, 
                conv_algorithm=args.conv_algorithm, 
                output_dir=args.output_dir,
                test_dataset_file=args.test_dataset_file,
                normalize=args.normalize)

def acetone_compare():
    parser = argparse.ArgumentParser(description='Program to verify the semantic preservation of ')

    parser.add_argument("reference_file", help="File with the inference output of the reference machine learning framework")
    parser.add_argument("c_file", help="File with the inference output of the studied machine learning framework")
    parser.add_argument("nb_tests", help="Number of inferences process to compare")
    parser.add_argument("--precision", help="Precision of the data studied. Default is float32")

    args = parser.parse_args()

    if args.precision:
        precision = args.precision
    else:
        precision = 'float'

    cli_compare(reference_file=args.reference_file,
                c_file=args.c_file,
                nb_tests=args.nb_tests,
                precision=precision)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from . import code_generator
from .code_generator import (
    CodeGenerator, Layer,
    Add_Bias, Concatenate, Dense, Dot, Flatten, Gather, Gemm, InputLayer, MatMul, Softmax, BatchNormalization,
    Add, Average, Broadcast, Divide, Maximum, Minimum, Multiply, Subtract,
    Conv2D, Conv2D_6loops, Conv2D_gemm, Conv2D_indirect_gemm, Conv2D_std_gemm,
    Pad, Edge_pad, Wrap_pad, Reflect_pad, Constant_Pad,
    Pooling2D, MaxPooling2D, AveragePooling2D,
    Resize, ResizeCubic, ResizeLinear, ResizeNearest,
    ActivationFunctions, Linear, Sigmoid, ReLu, TanH, Exponential, Logarithm, Clip
)

from .cli_acetone import cli_acetone
from .cli_compare import cli_compare

__all__ = (
    "cli_acetone", "cli_compare",
    code_generator.__all__
)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
import numpy as np
import pystache

#extract a list of subtensor from a given tensor
#attribut: axis alongside of which the submatrix will be extracted (if the desired submatrix must have the height, width or channels of the parent tensor)
#input: a tensor
#output: a list of tensor
class Gather(Layer):
    
    def __init__(self, idx, size, axis,  indices, input_shape, output_shape,activation_function):
        
        super().__init__()
        self.idx = idx
        self.size = size
        self.name = 'Gather'
        self.indices = indices
        self.axis = axis
        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]
        self.activation_function = activation_function
        
    def generate_inference_code_layer(self):
        output_str = self.previous_layer[0].output_str

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['output_str'] = output_str
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[position]')

        mustach_hash['indices_len'] = len(self.indices.flatten())
        mustach_hash['input_width'] = self.input_width
        mustach_hash['input_height'] = self.input_height

        if(self.axis == 1):
            mustach_hash['channels'] = True
            mustach_hash['output_height'] = self.output_height
            mustach_hash['output_width'] = self.output_width
        elif(self.axis == 2):
            mustach_hash['heights'] = True
            mustach_hash['output_channels'] = self.input_channels
            mustach_hash['output_width'] = self.output_width
        elif(self.axis == 3):
            mustach_hash['widths'] = True
            mustach_hash['output_channels'] = self.input_channels
            mustach_hash['output_height'] = self.output_height

        if(self.activation_function.name == 'linear'):
            mustach_hash['linear'] = True

        if(self.fused_layer):
            mustach_hash['fused_layer'] = self.fused_layer.write_activation_str('tensor_temp[position]',self.idx,'position')

        with open(self.template_path+'layers/template_Gather.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
        
    def forward_path_layer(self,input):
        input = input.reshape(self.input_channels,self.input_height,self.input_width)
        return np.take(input, indices=self.indices, axis=self.axis-1)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ...Layer import Layer

import numpy as np
from abc import abstractmethod

class Conv2D(Layer):
    
    def __init__(self, idx, conv_algorithm, size, padding, strides, kernel_h, kernel_w, dilation_rate, nb_filters, input_shape, output_shape, weights, biases, activation_function):
        super().__init__()
        self.conv_algorithm = conv_algorithm
        self.idx = idx
        self.size = size
        self.name = 'Conv2D'
        self.padding = padding
        self.strides = strides
        self.kernel_h = kernel_h
        self.kernel_w = kernel_w
        self.dilation_rate = dilation_rate
        self.nb_filters = nb_filters
    
        self.input_channels = input_shape[1]
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]

        self.input_shape = [self.input_channels, self.input_height, self.input_width]
        self.output_channels = self.nb_filters

        self.weights = np.asarray(weights)
        self.biases = np.asarray(biases)
        self.activation_function = activation_function
        self.local_var = 'sum'

        self.nb_weights = self.count_elements_array(self.weights)
        self.nb_biases = self.count_elements_array(self.biases)
        self.pad_right, self.pad_left, self.pad_bottom, self.pad_top = self.compute_padding(self.padding,self.input_height, self.input_width, self.kernel_h, self.kernel_w, self.strides, self.dilation_rate)
    
    @abstractmethod
    def generate_inference_code_layer(self):
        pass

    def forward_path_layer(self, input):
        # Conv for chw
        input = input.reshape(self.input_channels, self.input_height, self.input_width)
        
        output = np.zeros((self.nb_filters, self.output_height, self.output_width))
        print(self.weights.shape)

        if self.pad_right and self.pad_left and self.pad_top and self.pad_bottom:
            input_padded = np.zeros((self.input_channels, self.input_height + self.pad_top + self.pad_bottom, self.input_width + self.pad_left + self.pad_right))
            input_padded[:, self.pad_top:-self.pad_bottom, self.pad_left:-self.pad_right] = input
        else:
            input_padded = input
            
        for f in range(self.nb_filters):
            for i in range(self.output_height):
                for j in range(self.output_width):
                        output[f,i,j]=np.sum(input_padded[:, i*self.strides:i*self.strides+self.kernel_h, j*self.strides:j*self.strides+self.kernel_w] 
                                            * self.weights[:,:,:,f]) + self.biases[f]
        return self.activation_function.compute(output)


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
import numpy as np
import pystache


#Do the dotproduct of two tensors
#atribut: the axis alongside of which the dot product will be done. if type(axis) == int, same axis for the two tensor. can be tuples of axis (element i of axis represent axis of tensor i)
#input: two tensor
#output: the resultant tensor 
class Dot(Layer):
    def __init__(self, idx, size, axis, input_shapes,output_shape,activation_function):
        super().__init__()
        self.idx = idx
        self.size = size
        self.input_shapes = input_shapes
        self.name = 'Dot'
        #we seek a tuple of axis
        if (type(axis) == int):
            self.axis = [axis,axis]
        else:
            self.axis = axis
        self.output_fourth_dim = output_shape[1]
        self.output_height = output_shape[3]
        self.output_width = output_shape[4]
        self.output_channels = output_shape[2]
        self.activation_function = activation_function            

    def generate_inference_code_layer(self):

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['size'] = self.size
        mustach_hash['road'] = self.path

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('output')

        mustach_hash['output_fourth_dim'] = self.output_fourth_dim
        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width
        mustach_hash['axis_dim'] = self.input_shapes[0][self.axis[0]]
        mustach_hash['output_str_left'] = self.previous_layer[0].output_str
        mustach_hash['output_str_right'] = self.previous_layer[1].output_str

        if(self.axis[0] == 2):
            mustach_hash['indice_left'] = 'k + '+ str(self.input_shapes[0][self.axis[0]]) + ' * (f + ' + str(self.output_channels) + ' * p)'
        elif(self.axis[0] == 1):
            mustach_hash['indice_left'] = 'f + '+ str(self.output_channels) + ' * (k + ' + str(self.input_shapes[0][self.axis[0]]) + ' * p)'
        elif(self.axis[0] == 0):
            mustach_hash['indice_left'] = 'f + '+ str(self.output_channels) + ' * (p + ' + str(self.output_fourth_dim) + ' * k)'
        
        if(self.axis[1] == 2):
            mustach_hash['indice_right'] = 'k + '+ str(self.input_shapes[1][self.axis[1]]) + ' * (j + ' + str(self.output_width) + ' * i)'
        elif(self.axis[1] == 1):
            mustach_hash['indice_right'] = 'j + '+ str(self.output_width) + ' * (k + ' + str(self.input_shapes[1][self.axis[1]]) + ' * i)'
        elif(self.axis[1] == 0):
            mustach_hash['indice_right'] = 'j + '+ str(self.output_width) + ' * (i + ' + str(self.output_height) + ' * k)'
        
        with open(self.template_path+'layers/template_Dot.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def forward_path_layer(self, inputs):
        inputs[0] = inputs[0].reshape(self.input_shapes[0][1],self.input_shapes[0][2],self.input_shapes[0][3])
        inputs[1] = inputs[1].reshape(self.input_shapes[1][1],self.input_shapes[1][2],self.input_shapes[1][3])
        output = np.tensordot(inputs[0],inputs[1],axes=[self.axis[0]-1,self.axis[1]-1])
        return self.activation_function.compute(output)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""


import keras.backend
import numpy as np

from . import nnet_normalize

from ...code_generator.layers import Dense, InputLayer
from ...code_generator.activation_functions import Linear, ReLu

keras.backend.set_floatx('float32')


def load_nnet(file_to_parse, normalize):
    """
    Reads nnet networks from ACAS project and to pull it in .h5 format
    Inspired from : # https://github.com/NeuralNetworkVerification/Marabou/blob/master/maraboupy/MarabouNetworkNNet.py  read function
    Args:
        model_file_net (.nnet): file describing the neural network
    Returns: 
        model (.h5): Keras model
    """

    #Recall : Example of nnet file head:
    #7,5,5,50,  numLayers, inputSize, outputSize, maxLayersize
    #5,50,50,50,50,50,50,5,    Layersizes
    #0
    #Inputs Values Min
    #Inputs Values Max
    #Moyennes de normalisation
    #ranges de normalisation 
    #weights
    #biaises

    layers = []
    listRoad = [0,1]
    maxRoad = 1
    dict_cst = {}
    data_type = 'float'
    data_type_py = 'float32'

    with open(file_to_parse) as f:
     
        line = f.readline()

        while line[0:2] == "//": # ignore header lines with credits
            line = f.readline() 
                       
        # numLayers does't include the input layer!
        numLayers, inputSize, outputSize, maxLayersize = [int(x) for x in line.strip().split(",")[:-1]]
        line = f.readline()
        
        # input layer size, layer1size, layer2size...
        layerSizes = [int(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        # symmetric = int(line.strip().split(",")[0])

        line = f.readline()
        inputMinimums = [float(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        inputMaximums = [float(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        means = [float(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        ranges = [float(x) for x in line.strip().split(",")[:-1]]

        weights = []
        biases = []

        for layernum in range(numLayers):
            previousLayerSize = layerSizes[layernum]
            currentLayerSize = layerSizes[layernum + 1]
            weights.append([])
            biases.append([])
            
            # weights
            # for i in range(previousLayerSize):
            #     line = f.readline()
            #     aux = [float(x) for x in line.strip().split(",")[:-1]]
            #     weights[layernum].append([])
            #     for j in range(currentLayerSize):
            #         weights[layernum][i].append(aux[j])

            # weights for non conventional nnet
            for i in range(currentLayerSize):
                line = f.readline()
                aux = [float(x) for x in line.strip().split(",")[:-1]]
                weights[layernum].append([])
                for j in range(previousLayerSize):
                    weights[layernum][i].append(aux[j])
                    
            # biases
            for i in range(currentLayerSize):
                line = f.readline()
                x = float(line.strip().split(",")[0])
                biases[layernum].append(x)

    f.close()

    layers.append(InputLayer(idx=0, size=layerSizes[0], input_shape=[layerSizes[0]], data_format='channels_first'))
    layers[0].path = 0
    layers[0].find_output_str(dict_cst)
    
    # list of lists -> list of arrays
    weights = [np.transpose(np.array(weight)) for weight in weights]
    biases = [np.transpose(np.array(bias)) for bias in biases]

    for i in range(0,numLayers):
        weight = weights[i]
        if(len(weight.shape) < 4):
            for j in range(4-len(weight.shape)): 
                weight = np.expand_dims(weight, axis=0)

        if(i == numLayers-1):
            layer = Dense(idx=i+1,
                        size=layerSizes[i+1],
                        weights=weight,
                        biases=biases[i],
                        activation_function=Linear())
        else:
            layer = Dense(idx=i+1,
                        size=layerSizes[i+1],
                        weights=weight,
                        biases=biases[i],
                        activation_function=ReLu())
        
        layer.path = 0
        layer.find_output_str(dict_cst)
        layer.previous_layer.append(layers[-1])
        layers[-1].next_layer.append(layer)
        layers.append(layer)

    data_format = 'channels_first'

    print("Finished model initialization.")    

    if(normalize):
        Normalizer = nnet_normalize.Normalizer(input_size = layerSizes[0], output_size = layerSizes[-1], mins = inputMinimums, maxes = inputMaximums, means = means, ranges = ranges)
        return layers, data_type, data_type_py, data_format, maxRoad, dict_cst, Normalizer
    else:
        return layers, data_type, data_type_py, data_format, maxRoad, dict_cst


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import keras
from keras.engine.functional import Functional
from keras.engine.sequential import Sequential
import onnx

from .JSON_importer.parser_JSON import load_json
from .ONNX_importer.parser_ONNX import load_onnx
from .NNET_importer.parser_NNET import load_nnet
from .H5_importer.JSON_from_keras_model import JSON_from_keras_model
from .H5_importer.parser_h5 import load_keras

def get_path(file, new_type):
    new_path = ""
    list_path = file.split("/")
    file_name = list_path[-1].split(".")[0]

    for dir in list_path[:-1]:
        new_path += dir+"/"
    
    new_path += file_name + "." + new_type
    return new_path

def parser(file_to_parse, conv_algorithm, normalize=False):

    if(type(file_to_parse) == str):
        if("json" in  file_to_parse[-4:]):
            return load_json(file_to_parse, conv_algorithm)
        
        elif("onnx" in file_to_parse[-4:]):
            return load_onnx(file_to_parse, conv_algorithm)
        
        elif("h5" in file_to_parse[-4:]):
            return load_keras(file_to_parse, conv_algorithm)
        
        elif("nnet" in file_to_parse[-4:]):
            return load_nnet(file_to_parse,normalize)
        
        else:
            print("\nError: model description ."+file_to_parse[-4:]+" not supported")
            raise TypeError("Error: model description ."+file_to_parse[-4:]+" not supported\nOnly description supported are: .nnet, .h5, .json, .onnx\n")
    
    elif(type(file_to_parse) == onnx.ModelProto):
        return load_onnx(file_to_parse, conv_algorithm)
    
    elif(type(file_to_parse) == Functional or type(file_to_parse) == Sequential):
        return load_keras(file_to_parse, conv_algorithm)

    else:
        print("\nError: model description .",type(file_to_parse),"not supported")
        raise TypeError("Error: model description .",type(file_to_parse),"not supported\nOnly description supported are: .nnet, .h5, .json, .onnx\n")


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import os
import numpy as np
from pathlib import Path
from abc import ABC
import pystache

from .. import templates

from ..format_importer.parser import parser

from .layers import (
    Dense, Dot, Softmax, Gather, Gemm, MatMul, Concatenate, Pad, Broadcast, BatchNormalization,
    ResizeLinear, ResizeCubic, ResizeNearest, 
    Conv2D, Conv2D_6loops, Conv2D_indirect_gemm, Conv2D_std_gemm,
    Pooling2D, MaxPooling2D, AveragePooling2D
)

class CodeGenerator(ABC):

    def __init__(self, file, test_dataset_file = None, function_name = 'inference', nb_tests = None, conv_algorithm = 'conv_gemm_optim', normalize = False,**kwargs):

        self.file = file
        self.test_dataset_file = test_dataset_file
        self.function_name = function_name
        self.nb_tests = nb_tests
        self.conv_algorithm = conv_algorithm
        self.normalize = normalize
        self.template_path = templates.__file__[:-11]

        if (not self.normalize):
            l, dtype, dtype_py, data_format, maxpath, dict_cst = parser(self.file, self.conv_algorithm)
        elif(self.normalize):
            l, dtype, dtype_py, data_format, maxpath, dict_cst, Normalizer = parser(self.file, self.conv_algorithm, self.normalize)
            self.Normalizer = Normalizer
        
        self.layers = l
        self.data_type = dtype
        self.data_type_py = dtype_py
        self.maxpath = maxpath
        self.data_format = data_format
        self.dict_cst = dict_cst

        if test_dataset_file:
            ds = self.load_test_dataset()
            self.test_dataset = ds
        else:
            print("creating random dataset")
            ds = self.create_test_dataset()
            self.test_dataset = ds
            

        self.files_to_gen = ['inference.c', 'inference.h', 'global_vars.c', 'main.c', 'Makefile', 'test_dataset.h', 'test_dataset.c']
        
    def create_test_dataset(self):
        test_dataset = np.float32(np.random.default_rng(seed=10).random((int(self.nb_tests),1,int(self.layers[0].size))))
        return test_dataset 
     
    def load_test_dataset(self):
        
        test_dataset = []
        try:
            with open(self.test_dataset_file, 'r') as f:
                for i, line in enumerate(f):
                    line = line[1:-2].split(',')
                    if self.data_type == 'int':
                        line = list(map(int,line))
                    elif self.data_type == 'double':
                        line = list(map(float,line)) 
                    elif self.data_type == 'float':
                        line = list(map(np.float32,line))
                    test_dataset.append(line)
                    if i == int(self.nb_tests)-1:
                        break
            test_dataset = np.array(test_dataset)
            f.close()
       
        except TypeError:
            None
        
        return test_dataset

    def compute_inference(self, c_files_directory):
        with open(os.path.join(c_files_directory, 'output_python.txt'), 'w+') as fi:
            for nn_input in self.test_dataset:

                if((self.data_format == 'channels_last') and (len(self.layers[0].input_shape) == 4)): 
                    shape = (self.layers[0].input_shape[2],self.layers[0].input_shape[3],self.layers[0].input_shape[1])
                    nn_input = np.transpose(np.reshape(nn_input, shape), (2,0,1))

                if (self.normalize): nn_input = self.Normalizer.pre_processing(nn_input)
                previous_layer_result = [nn_input for i in range(self.maxpath)]  # for the very first layer, it is the neural network input
                
                to_store = {} #a dictionnary containing the values to store
                for layer in self.layers:
                    if(not layer.previous_layer):
                        previous_layer_result[layer.path] = layer.forward_path_layer(previous_layer_result[layer.path]) #if the layer is an input layer, it directly take the vaue from it's path
                    elif(len(layer.previous_layer)==1):
                        if(len(layer.previous_layer[0].next_layer) == 1):
                            previous_layer_result[layer.path] = layer.forward_path_layer(previous_layer_result[layer.previous_layer[0].path]) #if the layer has exactly one previous_layer, it takes the value from it's father's path
                        else:
                            previous_layer_result[layer.path] = layer.forward_path_layer(to_store[layer.previous_layer[0].idx]) #if the father is stored, we take it from the storage
                            layer.previous_layer[0].sorted +=1 #the number of children already "taken care of"
                    else:#if the layer has multiple ancestors, we take all of their value
                        prev_layer = []
                        for prev in layer.previous_layer:
                            if(len(prev.next_layer) == 1):
                                prev_layer.append(previous_layer_result[prev.path])
                            else:
                                prev_layer.append(to_store[prev.idx])
                                prev.sorted +=1
                        previous_layer_result[layer.path] = layer.forward_path_layer(prev_layer)
                    
                    #After having computed the value of the layer, we check if there is a fused layer.
                    if(layer.fused_layer):
                        #If the current layer is the last layer to be updated, the fused layer must be computed
                        if((layer.fused_layer.count_updated_prior_layers == len(layer.fused_layer.prior_layers)-1)):
                            fused = layer.fused_layer
                            #The layer has multiple ancestors (otherwise, it is treated as an activation function)
                            prev_layer = []
                            for prev in fused.prior_layers:
                                if(prev.idx not in to_store):# Taking the value where it is stored
                                    prev_layer.append(previous_layer_result[prev.path])
                                else:
                                    prev_layer.append(to_store[prev.idx])
                                    prev.sorted +=1
                            previous_layer_result[layer.path] = fused.forward_path_layer(prev_layer)
                        #Else, we notify the fused layer that another one of its ancestors have been computed
                        else:
                            layer.fused_layer.count_updated_prior_layers+=1
                    
                    if (len(layer.next_layer) > 1): #if the layer has more than one child, it needs to be stored
                        to_store[layer.idx] = previous_layer_result[layer.path]
                        
                    for prev in layer.previous_layer:
                        if ((prev.sorted == len(prev.next_layer)) and (prev in to_store)):#if all the children of the parent layer are "taken care of", we "forget" the parent's value ( *2 because of the creation of the dict in graph.to_save)
                            to_store.pop(prev.idx)
                            
                nn_output = previous_layer_result[layer.path]
                # print(nn_output) # write to file instead
                
                # Write results in text files to compare prediction.
                if((self.data_format == 'channels_last') and hasattr(layer, 'output_channels')): nn_output = np.transpose(nn_output, (1,2,0))
                nn_output = np.reshape(nn_output, -1)

                if(self.normalize):
                    nn_output = self.Normalizer.post_processing(nn_output)
                
                for j in range(len(nn_output)):
                    print('{:.9g}'.format(nn_output[j]), end=' ', file=fi, flush=True)           
                    # print(decimal.Decimal(nn_output[j]), end=' ', file=fi, flush=True)
                print(" ",file=fi)
        
        fi.close()

        print("File output_python.txt generated.")

        return nn_output

    def flatten_array_orderc(self, array):
    
        flattened_aray = array.flatten(order='C')
        s = '\n        {'
        for i in range(flattened_aray.size):
            s += str(flattened_aray[i])+', '
        s = s[:-2]
        s+='}'
        
        return s
    
    def flatten_array(self,array):
        s = '\n        {'
        shape = array.shape
        if(len(shape)<4):
            for i in range(4-len(shape)):
                shape = (1,) + shape
            array = np.reshape(array,shape)
        for j in range(shape[3]):
            for k in range(shape[0]):
                for f in range(shape[1]):
                    for i in range(shape[2]):
                        s+= str(array[k,f,i,j])+', '
        s = s[:-2]
        s+='}'
        return s

    def generate_testdataset_files(self):

        testdataset_header = open(self.c_files_directory + '/test_dataset.h' , "w+")
        testdataset_source = open(self.c_files_directory + '/test_dataset.c' , "w+")

        with open(self.template_path+'template_test_dataset_header.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        testdataset_header.write(pystache.render(template, {'nb_tests':self.nb_tests, 'nb_inputs':self.layers[0].size, 'nb_outputs':self.layers[-1].size, 'data_type':self.data_type}))
        testdataset_header.close()

        dataset = '{'
        if self.test_dataset is None:
            pass
        else:
            for j in range(self.test_dataset.shape[0]):
                dataset += self.flatten_array_orderc(self.test_dataset[j]) +','
            dataset = dataset[:-1]
        
        dataset += '};\n'

        with open(self.template_path+'template_test_dataset_source.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        testdataset_source.write(pystache.render(template,{'data_type':self.data_type, 'dataset':dataset}))
        testdataset_source.close()

    def generate_main_file(self):

        with open(self.template_path+'template_main_file.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        self.main_file.write(pystache.render(template, {'data_type':self.data_type}))
        self.main_file.close()

    def generate_makefile(self):

        header_files = []
        source_files = []
        for filename in self.files_to_gen:
            if '.c' in filename : source_files.append(filename)
            elif '.h' in filename : header_files.append(filename)
            else : pass

        with open(self.template_path+'template_Makefile.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        self.makefile.write(pystache.render(template, {'source_files':' '.join(source_files), 'header_files':' '.join(header_files), 'function_name':self.function_name}))
        self.makefile.close()

    def generate_c_files(self, c_files_directory):  

        self.c_files_directory = c_files_directory
        
        q = 0
        for file in self.files_to_gen:
            filename = Path(os.path.join(self.c_files_directory, file))

            if filename.is_file():
                print("ERROR : " + file + " already exists !")
                q += 1
            
        if q != 0 : exit()
        
        else:                               
            self.source_file = open(self.c_files_directory + '/inference.c' , "a+")
            self.header_file = open(self.c_files_directory + '/inference.h' , "a+")
            self.globalvars_file = open(self.c_files_directory + '/global_vars.c' , "a+")
            self.main_file = open(self.c_files_directory + '/main.c' , "a+")
            self.makefile = open(self.c_files_directory + '/Makefile' , "a+")

        self.generate_function_source_file()
        print('Generated function source file.') 
        self.generate_function_header_file()
        print('Generated function header file.') 
        self.generate_globalvars_file()
        print('Generated globalvars .c file.') 
        self.generate_main_file()
        print('Generated main file.')
        self.generate_makefile()
        print('Generated Makefile.')
        self.generate_testdataset_files()
        print('Generated testdataset files.')
      
    def generate_function_source_file(self):

        mustach_hash = {}

        mustach_hash['data_type'] = self.data_type
        mustach_hash['input_size'] = self.layers[0].size
        mustach_hash['output_size'] = self.layers[-1].size

        if any((isinstance(layer, Gather)) for layer in self.layers):
            mustach_hash['is_gather'] = True
            indices = []
            for gather in [layer for layer in self.layers if (isinstance(layer,Gather))]:
                indices.append({'idx':"{:02d}".format(gather.idx), 'lenght':len(gather.indices.flatten()), 'list':self.flatten_array_orderc(gather.indices)})
            mustach_hash['indices'] = indices

        self.l_size_max = 1
        for layer in (self.layers):
            if layer.size > self.l_size_max : self.l_size_max = layer.size
        
        if any((isinstance(layer, Dot) or isinstance(layer,Pooling2D) or isinstance(layer,Conv2D) or isinstance(layer,Gemm)) for layer in self.layers):
            mustach_hash['p'] = True

        if any((isinstance(layer, Conv2D_6loops) or isinstance(layer,Conv2D_std_gemm) or isinstance(layer,Pooling2D) or isinstance(layer,Gemm)) for layer in self.layers):
            mustach_hash['hw'] = True

        if any((isinstance(layer, Dense) or isinstance(layer,MatMul)) for layer in self.layers):
            mustach_hash['is_dense'] = True
        
        if (any(isinstance(layer, Conv2D_6loops) or isinstance(layer, AveragePooling2D) or isinstance(layer, Softmax)) for layer in self.layers):
            mustach_hash['is_sum'] = True
        
        if any(isinstance(layer, MaxPooling2D) for layer in self.layers):
            mustach_hash['is_max'] = True
               
        if any(isinstance(layer, AveragePooling2D) for layer in self.layers):
            mustach_hash['is_count'] = True
            
        if any(isinstance(layer,ResizeLinear) or isinstance(layer,ResizeCubic) or isinstance(layer,ResizeNearest) for layer in self.layers):
            mustach_hash['is_resize'] = True
        
        if any(isinstance(layer, ResizeCubic) for layer in self.layers):
            mustach_hash['is_cubic_interpolation'] = True
            
        if any(isinstance(layer, ResizeLinear) for layer in self.layers):
            mustach_hash['is_linear_interpolation'] = True
        
        mustach_hash['layers'] = []
        for layer in self.layers:
            layer_hash = {'inference_function':layer.generate_inference_code_layer(), 'path':layer.path, 'size':layer.size}
            
            if(layer in self.dict_cst):
                layer_hash['cst'] = True
                layer_hash['cst_name'] = self.dict_cst[layer]

            mustach_hash['layers'].append(layer_hash)
        
        output_hash = {}
        output_hash['path'] = self.layers[-1].path
        if ((self.data_format == 'channels_last') and (hasattr(self.layers[-1],'output_channels'))):
            output_hash['output_channels'] = self.layers[-1].output_channels
            output_hash['output_height'] = self.layers[-1].output_height
            output_hash['output_width'] = self.layers[-1].output_width

            with open(self.template_path+'memory_layout/template_channels_last_output.c.tpl','r') as template_file:
                template = template_file.read()
            template_file.close()
            mustach_hash['ouput_str'] = pystache.render(template, output_hash)

        else:
            output_hash['output_size'] = self.layers[-1].size
            if((self.data_format == 'channels_first')):
                output_hash['comment'] = 'Returning the output in channels first (ACETONE compute the result in channels first)'
            else:
                output_hash['comment'] = 'Returning the output (output flatten)'
            
            with open(self.template_path+'memory_layout/template_channels_first_output.c.tpl','r') as template_file:
                template = template_file.read()
            template_file.close()
            mustach_hash['ouput_str'] = pystache.render(template, output_hash)
            


        with open(self.template_path+'template_source_file.c.tpl', 'r') as template_file:
            template = template_file.read()
        template_file.close()

        if(self.normalize):
            mustach_hash['pre_processing'] = self.Normalizer.write_pre_processing()
            mustach_hash['post_processing'] = self.Normalizer.write_post_processing()
        
        self.source_file.write(pystache.render(template,mustach_hash))
        self.source_file.close()

    def generate_function_header_file(self):

        mustach_hash = {}
        mustach_hash['data_type'] = self.data_type
        mustach_hash['path'] = [i for i in range(self.maxpath)]

        self.nb_weights_max = 1
        self.nb_biases_max = 1

        self.patches_size_max = 1
        self.concate_size_max = 0
        for layer in self.layers: 
            if isinstance(layer, Conv2D_std_gemm):
                if layer.patches_size > self.patches_size_max : self.patches_size_max = layer.patches_size
            if isinstance(layer,Concatenate):
                self.patches_size_max = max(self.patches_size_max,layer.size)
                
        if any(isinstance(layer, Conv2D_std_gemm) for layer in self.layers):
            mustach_hash['path_size'] = max(self.l_size_max,self.patches_size_max)          
        else:
            mustach_hash['path_size'] = self.l_size_max
            
        mustach_hash['cst'] = []
        written = {}
        for layer in self.dict_cst:
            if self.dict_cst[layer] not in written:
                written[self.dict_cst[layer]] = layer.size
            else:
                written[self.dict_cst[layer]] = max(written[self.dict_cst[layer]],layer.size)

        for cst in written:
            mustach_hash['cst'].append({'name':cst, 'size':written[cst]})
            
        if (any(isinstance(layer, Concatenate) 
                or any(isinstance(layer, Conv2D)) 
                or any(isinstance(layer, Dense))
                or any(issubclass(layer, Broadcast)) 
                or any(isinstance(layer, Gather)) 
                or any(isinstance(layer, Pad))) for layer in self.layers):
            mustach_hash['tensor_temp'] = True
            mustach_hash['temp_size'] = max(self.l_size_max,self.patches_size_max)
 
        mustach_hash['layers'] = []
        for layer in self.layers:
            to_print = False
            layer_hash = {'name':layer.name, 'idx':"{:02d}".format(layer.idx)}

            if hasattr(layer, 'weights'):
                layer_hash['nb_weights']=layer.nb_weights
                if layer.nb_weights > self.nb_weights_max : self.nb_weights_max = layer.nb_weights
                to_print = True

            if hasattr(layer,'biases'):
                layer_hash['nb_biases']=layer.nb_biases
                if layer.nb_biases > self.nb_biases_max : self.nb_biases_max = layer.nb_biases
                to_print =True

            if type(layer) is Conv2D_indirect_gemm:
                layer_hash['patches_size'] = layer.patches_size
                to_print = True
            
            if type(layer) is BatchNormalization:
                layer_hash['channels'] = layer.output_channels
                to_print = True
            
            if issubclass(type(layer), Broadcast):
                if(layer.constant is not None):
                    layer_hash['constant_size'] = layer.constant_size
                    to_print = True

            if (to_print):
                mustach_hash['layers'].append(layer_hash)
        
        if(self.normalize):
            mustach_hash['normalization_cst'] = self.Normalizer.write_normalization_cst_in_header_file()
        
        with open(self.template_path+'template_header_file.c.tpl', 'r') as template_file:
            template = template_file.read()
        template_file.close()

        self.header_file.write(pystache.render(template,mustach_hash))
        self.header_file.close()

    def generate_globalvars_file(self):

        mustach_hash = {}

        mustach_hash['data_type'] = self.data_type
        mustach_hash['path'] = [i for i in range(self.maxpath)]

        if any(isinstance(layer, Conv2D_std_gemm) for layer in self.layers):
            mustach_hash['path_size'] = max(self.l_size_max,self.patches_size_max)          
        else:
            mustach_hash['path_size'] = self.l_size_max
        
        mustach_hash['cst'] = []
        written = {}
        for layer in self.dict_cst:
            if self.dict_cst[layer] not in written:
                written[self.dict_cst[layer]] = layer.size
            else:
                written[self.dict_cst[layer]] = max(written[self.dict_cst[layer]],layer.size)
        
        for cst in written:
            mustach_hash['cst'].append({'name':cst, 'size':written[cst]})
        
        if (any(isinstance(layer, Concatenate) 
                or any(isinstance(layer, Conv2D)) 
                or any(isinstance(layer, Dense))
                or any(issubclass(layer, Broadcast)) 
                or any(isinstance(layer, Gather)) 
                or any(isinstance(layer, Pad))) for layer in self.layers):
            mustach_hash['tensor_temp'] = True
            mustach_hash['temp_size'] = max(self.l_size_max,self.patches_size_max)
             
        if any(isinstance(layer, Conv2D_indirect_gemm) for layer in self.layers):
            mustach_hash['zero'] = True
        
        mustach_hash['layers'] = []

        for layer in self.layers:
            to_print = False
            layer_hash = {'name':layer.name, 'idx':"{:02d}".format(layer.idx)}

            if hasattr(layer, 'weights'):
                layer_hash['nb_weights'] = layer.nb_weights
                layer_hash['weights'] = self.flatten_array(layer.weights)
                to_print = True

            if hasattr(layer,'biases'):
                layer_hash['nb_biases'] = layer.nb_biases
                layer_hash['biases'] = self.flatten_array_orderc(layer.biases)
                to_print =True

            if type(layer) is Conv2D_indirect_gemm:
                layer_hash['patches_size'] = layer.patches_size
                layer_hash['patches'] = layer.create_ppatches()
                to_print = True
            
            if type(layer) is BatchNormalization:
                layer_hash['channels'] = layer.output_channels
                layer_hash['mean'] = self.flatten_array_orderc(layer.mean)
                layer_hash['var'] = self.flatten_array_orderc(layer.var)
                layer_hash['scale'] = self.flatten_array_orderc(layer.scale)
                to_print = True
            
            if issubclass(type(layer), Broadcast):
                if(layer.constant is not None):
                    layer_hash['constant'] = self.flatten_array_orderc(layer.constant)
                    layer_hash['constant_size'] = layer.constant_size
                    to_print = True
            
            if (to_print):
                mustach_hash['layers'].append(layer_hash)
        
        with open(self.template_path+'template_global_var_file.c.tpl', 'r') as template_file:
            template = template_file.read()
        template_file.close()

        self.globalvars_file.write(pystache.render(template,mustach_hash))

        if(self.normalize):
            self.globalvars_file.write(self.Normalizer.write_normalization_cst_in_globalvars_file())
        
        self.globalvars_file.close()


"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

from ..Layer import Layer
import numpy as np
import pystache

#Concatenate two tensor alongside a given axis
#attribut: axis alongside of which the concatenation will be done
#input: a list of tensor to concatenate
#output: the concatenated tensor 
class Concatenate(Layer):
    def __init__(self, idx, size, axis, input_shapes,output_shape,activation_function):
        super().__init__()
        self.idx = idx
        self.size = size
        self.input_shapes = input_shapes
        self.name = 'Concatenate'
        self.axis = axis
        self.output_height = output_shape[2]
        self.output_width = output_shape[3]
        self.output_channels = output_shape[1]
        self.activation_function = activation_function            
    
    def generate_inference_code_layer(self):
        borne_sup = 0
        borne_inf = 0

        mustach_hash = {}

        mustach_hash['name'] = self.name
        mustach_hash['idx'] = "{:02d}".format(self.idx)
        mustach_hash['comment'] = self.activation_function.comment
        mustach_hash['road'] = self.path
        mustach_hash['size'] = self.size

        mustach_hash['activation_function'] = self.activation_function.write_activation_str('tensor_temp[k]')

        mustach_hash['output_channels'] = self.output_channels
        mustach_hash['output_height'] = self.output_height
        mustach_hash['output_width'] = self.output_width

        if(self.axis == 1):
            mustach_hash['channels'] = True
        elif(self.axis == 2):
            mustach_hash['heights'] = True
        elif(self.axis == 3):
            mustach_hash['widths'] = True

        mustach_hash['concat'] = []
        for k in range(len(self.previous_layer)):
            borne_sup += self.input_shapes[k][self.axis]

            layer_to_concat = {}
            layer_to_concat['input_width'] = self.input_shapes[k][3]
            layer_to_concat['input_height'] = self.input_shapes[k][2]
            layer_to_concat['output_str'] = self.previous_layer[k].output_str
            layer_to_concat['borne_sup'] = borne_sup
            layer_to_concat['borne_inf'] = borne_inf
            mustach_hash['concat'].append(layer_to_concat)

            borne_inf += self.input_shapes[k][self.axis]


        with open(self.template_path+'layers/template_Concatenate.c.tpl','r') as template_file:
            template = template_file.read()
        template_file.close()

        return pystache.render(template, mustach_hash)
    
    def forward_path_layer(self, inputs):
        output = inputs[0]
        output = output.reshape(self.input_shapes[0][1],self.input_shapes[0][2],self.input_shapes[0][3])
        for i in range(1,len(inputs)):
            input = inputs[i]
            input = input.reshape(self.input_shapes[i][1],self.input_shapes[i][2],self.input_shapes[i][3])
            output = np.concatenate((output, input),axis=self.axis - 1) 
        return self.activation_function.compute(output)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import sys
import argparse
import numpy as np

# Source document for floating comparison
# https://floating-point-gui.de/errors/comparison/


def compare_floats(a, b, epsilon = (128*sys.float_info.epsilon), abs_th = sys.float_info.min):
    diff = abs(a-b) 
    if a == b:
        return True, diff

    else:
        norm = min((abs(a)+abs(b)), (sys.float_info.max))
        if diff < max(abs_th, epsilon * norm):
            return True, diff
        else: 
            return False, diff

def preprocess_line(line, precision):
    line = line.split(' ')
    line[:] = [x for x in line if x.strip()]
    if precision == 'double':
        line = list(map(np.float64,line))
    elif precision == 'float':
        line = list(map(np.float32,line))  
    return line 

def compare_lines(line_f1, line_f2, precision):

    line_f1 = preprocess_line(line_f1, precision)
    line_f2 = preprocess_line(line_f2, precision)
    max_diff = 0
    length = len(line_f1)

    for j in range(length): 
        comparison, diff = compare_floats(line_f1[j],line_f2[j])
        if diff > max_diff:
            max_diff = diff

    if comparison:
        return True, max_diff
    else:
        return False, max_diff


def compare_files(file1, file2, nb_tests, precision):
    f1 = open(file1, 'r')
    f2 = open(file2, 'r')

    line_counter = 0
    max_diff_file = 0


    for (line_f1, line_f2) in zip(f1, f2):
        line_counter += 1
        comparison, max_diff_line = compare_lines(line_f1, line_f2, precision)
        
        if max_diff_line > max_diff_file:
            max_diff_file = max_diff_line

        if line_counter == int(nb_tests):
            break

    f1.close()
    f2.close()


    if comparison:
        return True, max_diff_file
        
    else: 
        return False, max_diff_file

def cli_compare(reference_file, c_file, nb_tests, precision):
    _, max_diff_file = compare_files(reference_file, c_file, nb_tests, precision)
    
    print("   Max absolute error for %s test(s): %s" % (nb_tests, max_diff_file))


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Program to verify the semantic preservation of ')

    parser.add_argument("reference_file", help="File with the inference output of the reference machine learning framework")
    parser.add_argument("c_file", help="File with the inference output of the studied machine learning framework")
    parser.add_argument("nb_tests", help="Number of inferences process to compare")
    parser.add_argument("--precision", help="Precision of the data studied. Default is float32")

    args = parser.parse_args()

    if args.precision:
        precision = args.precision
    else:
        precision = 'float'

    cli_compare(args.reference_file, args.c_file, args.nb_tests, precision)



"""
 *******************************************************************************
 * ACETONE: Predictable programming framework for ML applications in safety-critical systems
 * Copyright (c) 2022. ONERA
 * This file is part of ACETONE
 *
 * ACETONE is free software ;
 * you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation ;
 * either version 3 of  the License, or (at your option) any later version.
 *
 * ACETONE is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this program ;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 ******************************************************************************
"""

import numpy as np
from abc import ABC, abstractmethod
from .. import templates

class Layer(ABC):
    
    def __init__(self):

        self.idx = 0
        self.size = 0
        self.name = ''
        self.next_layer = [] 
        self.previous_layer = []
        self.path = None
        self.sorted = None
        self.output_str = ''
        self.fused_layer = None
        self.template_path = templates.__file__[:-11]
      
        super().__init__()

    @abstractmethod
    def generate_inference_code_layer(self):
        pass

    @abstractmethod
    def forward_path_layer(self):
        pass

    def count_elements_array(self, array):
        nb_elements = 1
        for dim in np.shape(array) : nb_elements *= dim
        return nb_elements

    def compute_padding(self,padding, in_height, in_width, kernel_h, kernel_w, strides, dilation_rate=1):
        if(type(padding) == str):
            # Compute 'same' padding tensorflow

            filter_height = (kernel_h - (kernel_h-1)*(dilation_rate-1))
            filter_width = (kernel_w - (kernel_w-1)*(dilation_rate-1))

            # The total padding applied along the height and width is computed as:
            if(padding == 'VALID' or padding == 'valid'):
                pad_right, pad_left, pad_bottom, pad_top = 0,0,0,0
            else:
                if (in_height % strides == 0):
                    pad_along_height = max(filter_height - strides, 0)
                else:
                    pad_along_height = max(filter_height - (in_height % strides), 0)
                if (in_width % strides == 0):
                    pad_along_width = max(filter_width - strides, 0)
                else:
                    pad_along_width = max(filter_width - (in_width % strides), 0)
                
                if((padding == 'SAME_LOWER') or (padding == 'same')):
                    pad_top = pad_along_height // 2
                    pad_bottom = pad_along_height - pad_top
                    pad_left = pad_along_width // 2
                    pad_right = pad_along_width - pad_left
                elif(padding == 'SAME_UPPER'):
                    pad_bottom = pad_along_height // 2
                    pad_top = pad_along_height - pad_bottom
                    pad_right = pad_along_width // 2
                    pad_left = pad_along_width - pad_right        
        else:
            pad_right, pad_left, pad_bottom, pad_top = padding[3], padding[1], padding[2], padding[0]
            
        return pad_right, pad_left, pad_bottom, pad_top
    
    #Give to the layer an string saying were the output will be saved (either in a 'cst' or in an 'output_road')
    def find_output_str(self,dict_cst):
        #dict_cst is the dict linking an layer to the cst in which the must be saved if needed
        
        #either it has to be saved
        if(self in dict_cst):
            output_str = 'cst_'+str(dict_cst[self])
        #Or it can directly go to the next layer
        else:
            output_str = 'output_'+str(self.path)
        self.output_str = output_str
        return self